import os
import json
import ujson
import random
import time
import pickle
import logger
import numpy as np
import torch
import math
import lightning as L
import torch.distributed as dist
from transformers import AutoModel
from pytorch_transformers.optimization import WarmupLinearSchedule
from datetime import datetime
from typing import Optional, Union

from scipy.sparse.csgraph import minimum_spanning_tree

# csgraph = compressed sparse graph
from scipy.sparse import csr_matrix

# csr_matrix = compressed sparse row matrices
from collections import Counter

import bioel.models.arboel.biencoder.data.data_process as data_process
import bioel.models.arboel.biencoder.model.eval_cluster_linking as eval_cluster_linking
from bioel.models.arboel.biencoder.model.biencoder import BiEncoderRanker
from bioel.models.arboel.biencoder.model.common.optimizer import get_bert_optimizer
from bioel.models.arboel.biencoder.model.common.params import BlinkParser
from bioel.models.arboel.biencoder.model.eval_cluster_linking import (
    filter_by_context_doc_id,
)
from bioel.models.arboel.biencoder.model.special_partition.special_partition import (
    cluster_linking_partition,
)
from bioel.utils.bigbio_utils import (
    load_bigbio_dataset,
    dataset_to_df,
    add_deabbreviations,
    CUIS_TO_REMAP,
    CUIS_TO_EXCLUDE,
)

from IPython import embed
from tqdm import tqdm

import logging

# Configure the logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

from bioel.models.arboel.biencoder.data.BiEncoderLightningDataModule import (
    ArboelDataModule,
)


def create_versioned_filename(base_path, base_name, extension=".json"):
    """Create a versioned filename to avoid overwriting existing files.
    Params:
    - base_path (str):
        The directory where the file will be saved.
    base_name (str):
        The base name of the file without the extension.
    extension (str):
        The file extension.
    Returns:
        str: A versioned filename that does not exist in the base path.
    """
    version = 1
    while True:
        new_filename = f"{base_name}{version}{extension}"
        full_path = os.path.join(base_path, new_filename)
        if not os.path.exists(full_path):
            return full_path
        version += 1


def evaluate(
    params,
    logger,
    entity_data,
    query_data,
    valid_men_embeds,
    nn_ent_dists,
    nn_ent_idxs,
    nn_men_dists,
    nn_men_idxs,
    n_entities,
    n_mentions,
    max_knn,
    use_types=False,
    within_doc=False,
    context_doc_ids=None,
):
    """
    Description
    -----------
    The first 2 steps are done are done at the beginning of each validation epoch.
    1) Computes embeddings and faiss indexes for entities and mentions.
    2) Performs k-nearest neighbors (k-NN) search to establish relationships between them.
    3) Constructs graphs based on these relationships.
    4) Evaluates the model's accuracy by analyzing how effectively the model can link mentions to the correct entities.

    Parameters
    ----------
    logger : 'Logger' object
        Logging object used to record messages
    entity_data : list of dict
        Contains informations (type, cui, description, tokens, ids) on entities
    query_data : list of dict
        Contains information about mentions (mention_id, mention_name, context, label_idxs, etc…) of validation set
    use_types=False : bool
        A boolean flag that indicates whether or not to use type-specific indexes for entities and mentions
    valid_men_embeds : tensor of dim = (nb of entities, entities length = nb of tokens = 64)
        Contains the embedding of the entity token ID vectors
    nn_ent_dists : float
        Distance of the closest entity for all mentions in "valid_men_embeds"
    nn_ent_idxs : int
        Indice of the closest entity for all mentions in "valid_men_embeds"
    nn_men_dists : list of float
        Distance of the k closest mentions for all mentions in "valid_men_embeds"
    nn_men_idxs : list of int
        Indice of the k closest mentions for all mentions in "valid_men_embeds"
    n_entities : int
        Total number of entities
    n_mentions : int
        Total number of mentions in the validation set.
    max_knn : int
        max number of knn
    within_doc=False : bool
        Boolean flag that indicates whether the evaluation should be constrained to within-document contexts
    context_doc_ids=None : bool
        This would be used in conjunction with within_doc to limit evaluations within the same document.
    """

    joint_graphs = (
        {}
    )  # Store results of the NN search and distance between entities and mentions

    knn_vals = [0] + [2**i for i in range(int(math.log(params["knn"], 2)) + 1)]

    for k in knn_vals:
        joint_graphs[k] = {
            "rows": np.array([]),
            "cols": np.array([]),
            "data": np.array([]),
            "shape": (n_entities + n_mentions, n_entities + n_mentions),
        }

    "3) Constructs graphs based on these relationships."
    """
    nn_ent_dists contain information about distance of the closest entity
    nn_ent_idxs contain information about indice of the closest entity
    nn_men_dists contain information about distance of the k nearest mentions
    nn_men_idxs contain information about indice of the k nearest mentions
    - We can fill in the "rows" part (=start nodes) of the graph in the order of the mentions
    - We can fill in the "cols" part (=end nodes) of the graph with nn_ent_idxs and nn_men_idxs
    - We can fill in the "data" part (=weights) of the graph with nn_ent_dists and nn_men_dists
    """
    logger.info("Eval: Building graphs")
    for men_query_idx, men_embed in enumerate(
        tqdm(
            valid_men_embeds, total=len(valid_men_embeds), desc="Eval: Building graphs"
        )
    ):
        # Get nearest entity candidate
        dict_cand_idx = nn_ent_idxs[men_query_idx][
            0
        ]  # Use of [0] to retrieve a scalar and not an 1D array
        dict_cand_score = nn_ent_dists[men_query_idx][0]

        # Filter candidates to remove -1s, mention query, within doc (if reqd.), and keep only the top k candidates
        filter_mask_neg1 = (
            nn_men_idxs[men_query_idx] != -1
        )  # bool ndarray. Ex : np.array([True, False, True, False])
        men_cand_idxs = nn_men_idxs[men_query_idx][
            filter_mask_neg1
        ]  # Only keep the elements != -1
        men_cand_scores = nn_men_dists[men_query_idx][filter_mask_neg1]

        if within_doc:
            men_cand_idxs, wd_mask = filter_by_context_doc_id(
                mention_idxs=men_cand_idxs,
                doc_id=context_doc_ids[men_query_idx],
                doc_id_list=context_doc_ids,
                return_numpy=True,
            )
            men_cand_scores = men_cand_scores[wd_mask]

        # Filter self-reference + limits the number of candidate to 'max_knn'
        filter_mask = men_cand_idxs != men_query_idx
        men_cand_idxs, men_cand_scores = (
            men_cand_idxs[filter_mask][:max_knn],
            men_cand_scores[filter_mask][:max_knn],
        )

        # Add edges to the graphs
        for k in joint_graphs:
            joint_graph = joint_graphs[k]
            # Add mention-entity edge
            joint_graph["rows"] = (
                np.append(  # Mentions are offset by the total number of entities to differentiate mention nodes from entity nodes
                    joint_graph["rows"], [n_entities + men_query_idx]
                )
            )
            joint_graph["cols"] = np.append(joint_graph["cols"], dict_cand_idx)
            joint_graph["data"] = np.append(joint_graph["data"], dict_cand_score)
            if k > 0:
                # Add mention-mention edges
                joint_graph["rows"] = np.append(
                    joint_graph["rows"],
                    [n_entities + men_query_idx]
                    * len(
                        men_cand_idxs[:k]
                    ),  # Creates an array where the starting node (current mention) is repeated len(men_cand_idxs[:k]) times
                )
                joint_graph["cols"] = np.append(
                    joint_graph["cols"], n_entities + men_cand_idxs[:k]
                )
                joint_graph["data"] = np.append(
                    joint_graph["data"], men_cand_scores[:k]
                )

    "4) Evaluates the model's accuracy by analyzing how effectively the model can link mentions to the correct entities."
    dict_acc = {}
    max_eval_acc = -1.0
    for k in joint_graphs:
        logger.info(f"\nEval: Graph (k={k}):")
        # Partition graph based on cluster-linking constraints (inference procedure)
        partitioned_graph, clusters = eval_cluster_linking.partition_graph(
            joint_graphs[k], n_entities, directed=True, return_clusters=True
        )
        # Infer predictions from clusters
        result = eval_cluster_linking.analyzeClusters(
            clusters, entity_data, query_data, k
        )
        acc = float(result["accuracy"].split(" ")[0])
        dict_acc[f"k{k}"] = acc
        max_eval_acc = max(acc, max_eval_acc)
        logger.info(f"Eval: accuracy for graph@k={k}: {acc}%")
    logger.info(f"Eval: Best accuracy: {max_eval_acc}%")
    return max_eval_acc, dict_acc


def evaluate_test(
    params,
    reranker,
    test_dict_vecs,
    test_processed_data,
    test_men_vecs,
    logger,
    entity_data,
    train_processed_data=None,
    train_men_vecs=None,
    use_types=False,
    embed_batch_size=768,
    force_exact_search=False,
    probe_mult_factor=1,
    within_doc=False,
    context_doc_ids=None,
):
    """
    Description : Specific evaluation for test set (can control the number of recall k + save data)
    -----------
    1) Computes embeddings and indexes for entities and mentions.
    2) Performs k-nearest neighbors (k-NN) search to establish relationships between them.
    3) Constructs graphs based on these relationships.
    4) Evaluates the model's accuracy by analyzing how effectively the model can link mentions to the correct entities.

    Parameters
    ----------
    reranker : BiEncoderRanker
        NN-based ranking model
    test_dict_vec : list of tensors
        Contains the IDs of the entities (title + description) tokens
    test_processed_data : list of dict
        Contains information about mentions (mention_id, mention_name, context, etc…) of test set
    test_men_vecs : list of tensors
        Contains only the IDs of (mention + surrounding context) tokens of test set
    logger : 'Logger' object
        Logging object used to record messages
    entity_data : list or dict
        Entities from the data
    train_processed_data : list of dict
        Contains information about mentions (mention_id, mention_name, context, etc…) of train set
    train_men_vecs : list of tensors
        Contains only the IDs of (mention + surrounding context) tokens of train set
    use_types=False : bool
        A boolean flag that indicates whether or not to use type-specific indexes for entities and mentions
    embed_batch_size=128 : int
        The batch size to use when processing embeddings.
    force_exact_search=False : bool
        force the embedding process to use exact search methods rather than approximate methods.
    probe_mult_factor=1 : int
        A multiplier factor used in index building for probing in case of approximate search (bigger = better but slower)
    within_doc=False : bool
        Boolean flag that indicates whether the evaluation should be constrained to within-document contexts
    context_doc_ids=None : bool
        This would be used in conjunction with within_doc to limit evaluations within the same document.
    """

    output_path = params["output_path"]
    if not os.path.exists(output_path):
        os.makedirs(output_path)

    embed_data_path = params["embed_data_path"]
    if embed_data_path is None or not os.path.exists(embed_data_path):
        embed_data_path = output_path

    knn = params["knn"]

    n_entities = len(test_dict_vecs)
    n_mentions = len(test_men_vecs)  ## Change

    if params["transductive"]:
        n_mentions += len(train_men_vecs)
        n_train_mentions = len(train_men_vecs)

    # Values of k to run the evaluation against
    knn_vals = [0] + [2**i for i in range(int(math.log(knn, 2)) + 1)]
    # Store the maximum evaluation k
    max_knn = knn_vals[-1]

    # Maps cui to unique identifier
    dict_cui_to_idx = {}
    for idx, ent in enumerate(entity_data):
        dict_cui_to_idx[ent["cui"]] = idx
    # Maps cui to unique identifier : Used for retrieving cui of top candidates
    dict_idx_to_cui = {v: k for k, v in dict_cui_to_idx.items()}

    # Needed for the output for evaluation
    data = load_bigbio_dataset(params["dataset"])
    if params["path_to_abbrev"]:
        data_with_abbrev = add_deabbreviations(
            dataset=data, path_to_abbrev=params["path_to_abbrev"]
        )
    exclude = CUIS_TO_EXCLUDE[params["dataset"]]
    remap = CUIS_TO_REMAP[params["dataset"]]
    df = dataset_to_df(
        data_with_abbrev, entity_remapping_dict=remap, cuis_to_exclude=exclude
    )

    # Output for evaluation object
    output_eval = []

    time_start = time.time()

    # Check if graphs are already built
    graph_path = os.path.join(output_path, "graphs.pickle")
    if not params["only_recall"] and os.path.isfile(graph_path):
        print("Loading stored joint graphs...")
        with open(graph_path, "rb") as read_handle:
            joint_graphs = pickle.load(read_handle)
    else:
        # Initialize graphs to store mention-mention and mention-entity similarity score edges;
        # Keyed on k, the number of nearest mentions retrieved
        joint_graphs = {}
        for k in knn_vals:
            joint_graphs[k] = {
                "rows": np.array([]),
                "cols": np.array([]),
                "data": np.array([]),
                "shape": (n_entities + n_mentions, n_entities + n_mentions),
            }
        # Check and load stored embedding data
        embed_data = None
        if os.path.isfile(embed_data_path):
            embed_data = torch.load(params["embed_data_path"])

        "1) Computes embeddings and indexes for entities and mentions. "
        if use_types:
            if embed_data is not None:
                logger.info("Dictionary : Loading stored embedding (entities)")
                dict_embeds = embed_data["dict_embeds"]
                if "dict_idxs_by_type" in embed_data:
                    dict_idxs_by_type = embed_data["dict_idxs_by_type"]
                else:
                    dict_idxs_by_type = data_process.get_idxs_by_type(entity_data)
                dict_indexes = data_process.get_index_from_embeds(
                    dict_embeds,
                    dict_idxs_by_type,
                    force_exact_search=params["force_exact_search"],
                    probe_mult_factor=params["probe_mult_factor"],
                )

                logger.info("Queries: Embedding and building index (mentions)")
                vecs = test_men_vecs  # entity_dict_vecs
                men_data = test_processed_data
                if params["transductive"]:
                    vecs = torch.cat((train_men_vecs, vecs), dim=0)
                    men_data = train_processed_data + test_processed_data
                men_embeds, men_indexes, men_idxs_by_type = (
                    data_process.embed_and_index(
                        model=reranker,
                        token_id_vecs=vecs,
                        encoder_type="context",
                        corpus=men_data,
                        force_exact_search=params["force_exact_search"],
                        batch_size=params["embed_batch_size"],
                        probe_mult_factor=params["probe_mult_factor"],
                    )
                )

            else:
                logger.info("Dictionary: Embedding and building index (entities)")
                dict_embeds, dict_indexes, dict_idxs_by_type = (
                    data_process.embed_and_index(
                        model=reranker,
                        token_id_vecs=test_dict_vecs,
                        encoder_type="candidate",
                        corpus=entity_data,
                        force_exact_search=params["force_exact_search"],
                        batch_size=params["embed_batch_size"],
                        probe_mult_factor=params["probe_mult_factor"],
                    )
                )

                logger.info("Queries: Embedding and building index (mentions)")
                vecs = test_men_vecs
                men_data = test_processed_data
                if params["transductive"]:
                    vecs = torch.cat((train_men_vecs, vecs), dim=0)
                    men_data = train_processed_data + test_processed_data
                men_embeds, men_indexes, men_idxs_by_type = (
                    data_process.embed_and_index(
                        model=reranker,
                        token_id_vecs=vecs,
                        encoder_type="context",
                        corpus=men_data,
                        force_exact_search=params["force_exact_search"],
                        batch_size=params["embed_batch_size"],
                        probe_mult_factor=params["probe_mult_factor"],
                    )
                )
        else:
            if embed_data is not None:
                logger.info("Dictionary : Loading stored embedding (entities)")
                dict_embeds = embed_data["dict_embeds"]
                dict_index = data_process.get_index_from_embeds(
                    dict_embeds,
                    force_exact_search=params["force_exact_search"],
                    probe_mult_factor=params["probe_mult_factor"],
                )

                logger.info("Queries: Embedding and building index (mentions)")
                vecs = test_men_vecs
                if params["transductive"]:
                    vecs = torch.cat((train_men_vecs, vecs), dim=0)
                men_embeds, men_index = data_process.embed_and_index(
                    model=reranker,
                    token_id_vecs=vecs,
                    encoder_type="context",
                    force_exact_search=params["force_exact_search"],
                    batch_size=params["embed_batch_size"],
                    probe_mult_factor=params["probe_mult_factor"],
                )
            else:
                logger.info("Dictionary: Embedding and building index (entities)")
                dict_embeds, dict_index = data_process.embed_and_index(
                    model=reranker,
                    token_id_vecs=test_dict_vecs,
                    encoder_type="candidate",
                    force_exact_search=params["force_exact_search"],
                    batch_size=params["embed_batch_size"],
                    probe_mult_factor=params["probe_mult_factor"],
                )

                logger.info("Queries: Embedding and building index (mentions)")
                vecs = test_men_vecs
                if params["transductive"]:
                    vecs = torch.cat((train_men_vecs, vecs), dim=0)
                men_embeds, men_index = data_process.embed_and_index(
                    model=reranker,
                    token_id_vecs=vecs,
                    encoder_type="context",
                    force_exact_search=params["force_exact_search"],
                    batch_size=params["embed_batch_size"],
                    probe_mult_factor=params["probe_mult_factor"],
                )

        recall_accuracy = {
            2**i: 0 for i in range(int(math.log(params["recall_k"], 2)) + 1)
        }
        recall_idxs = [0.0] * params["recall_k"]

        "2) Performs k-nearest neighbors (k-NN) search to establish relationships between mentions and entities."
        logger.info("Starting KNN search...")
        # Fetch recall_k (default 16) knn entities for all mentions
        if not use_types:
            _men_embeds = men_embeds
            if params["transductive"]:
                _men_embeds = _men_embeds[n_train_mentions:]
            nn_ent_dists, nn_ent_idxs = dict_index.search(
                _men_embeds, params["recall_k"]
            )
            n_mens_to_fetch = len(_men_embeds) if within_doc else max_knn + 1
            nn_men_dists, nn_men_idxs = men_index.search(_men_embeds, n_mens_to_fetch)
        else:
            query_len = len(men_embeds) - (
                n_train_mentions if params["transductive"] else 0
            )
            nn_ent_idxs = np.zeros((query_len, params["recall_k"]))
            nn_ent_dists = np.zeros((query_len, params["recall_k"]), dtype="float64")
            nn_men_idxs = -1 * np.ones((query_len, query_len), dtype=int)
            nn_men_dists = -1 * np.ones((query_len, query_len), dtype="float64")
            for entity_type in men_indexes:
                men_embeds_by_type = (
                    men_embeds[
                        men_idxs_by_type[entity_type][
                            men_idxs_by_type[entity_type] >= n_train_mentions
                        ]
                    ]
                    if params["transductive"]
                    else men_embeds[men_idxs_by_type[entity_type]]
                )
                nn_ent_dists_by_type, nn_ent_idxs_by_type = dict_indexes[
                    entity_type
                ].search(men_embeds_by_type, params["recall_k"])
                nn_ent_idxs_by_type = np.array(
                    list(
                        map(
                            lambda x: dict_idxs_by_type[entity_type][x],
                            nn_ent_idxs_by_type,
                        )
                    )
                )
                n_mens_to_fetch = len(men_embeds_by_type) if within_doc else max_knn + 1
                nn_men_dists_by_type, nn_men_idxs_by_type = men_indexes[
                    entity_type
                ].search(
                    men_embeds_by_type, min(n_mens_to_fetch, len(men_embeds_by_type))
                )
                nn_men_idxs_by_type = np.array(
                    list(
                        map(
                            lambda x: men_idxs_by_type[entity_type][x],
                            nn_men_idxs_by_type,
                        )
                    )
                )
                i = -1
                for idx in men_idxs_by_type[entity_type]:
                    if params["transductive"]:
                        idx -= n_train_mentions
                    if idx < 0:
                        continue
                    i += 1
                    nn_ent_idxs[idx] = nn_ent_idxs_by_type[i]
                    nn_ent_dists[idx] = nn_ent_dists_by_type[i]
                    nn_men_idxs[idx][: len(nn_men_idxs_by_type[i])] = (
                        nn_men_idxs_by_type[i]
                    )
                    nn_men_dists[idx][: len(nn_men_dists_by_type[i])] = (
                        nn_men_dists_by_type[i]
                    )
        logger.info("Search finished")

        "3) Constructs graphs based on these relationships."
        logger.info("Building graphs")
        # Find the most similar entity and k-nn mentions for each mention query
        for idx in range(len(nn_ent_idxs)):
            # Get nearest entity candidate
            dict_cand_idx = nn_ent_idxs[idx][0]

            # Indices of nearest entity candidates
            idx_top_candidates = nn_ent_idxs[idx][: params["recall_k"]]
            # Cuis of nearest entity candidates
            cuis_top_candidates = [
                [dict_idx_to_cui.get(index)] for index in idx_top_candidates
            ]

            mention_id = test_processed_data[idx]["mention_id"]
            mention_name = test_processed_data[idx]["mention_name"]

            # Check if the mention_id with .abbr_resolved exists
            if (df["mention_id"] + ".abbr_resolved" == mention_id).any():
                filtered_df = df[
                    (df["mention_id"] + ".abbr_resolved" == mention_id)
                ].copy()
                filtered_df["mention_id"] += ".abbr_resolved"
            else:
                filtered_df = df[df["mention_id"] == mention_id]

            if not filtered_df.empty:
                mention_dict = filtered_df.iloc[0].to_dict()
            else:
                print(
                    "This mention name was not found in the mention dataset:",
                    mention_id,
                    mention_name,
                )

            mention_dict["candidates"] = cuis_top_candidates

            if params["equivalent_cuis"]:
                synsets = ujson.load(
                    open(os.path.join(params["data_path"], "cui_synsets.json"))
                )
                mention_dict["candidates"] = [
                    synsets[y[0]] for y in mention_dict["candidates"]
                ]

            output_eval.append(mention_dict)

            dict_cand_score = nn_ent_dists[idx][0]
            # Compute recall metric
            gold_idxs = test_processed_data[idx]["label_idxs"][
                : test_processed_data[idx]["n_labels"]
            ]

            recall_idx = np.argwhere(nn_ent_idxs[idx] == gold_idxs[0])
            if len(recall_idx) != 0:
                recall_idxs[int(recall_idx)] += 1.0
                for recall_k in recall_accuracy:
                    if recall_idx < recall_k:
                        recall_accuracy[recall_k] += 1.0
            if not params["only_recall"]:
                filter_mask_neg1 = nn_men_idxs[idx] != -1
                men_cand_idxs = nn_men_idxs[idx][filter_mask_neg1]
                men_cand_scores = nn_men_dists[idx][filter_mask_neg1]

                if within_doc:
                    men_cand_idxs, wd_mask = filter_by_context_doc_id(
                        men_cand_idxs,
                        context_doc_ids[idx],
                        context_doc_ids,
                        return_numpy=True,
                    )
                    men_cand_scores = men_cand_scores[wd_mask]

                # Filter candidates to remove mention query and keep only the top k candidates
                filter_mask = men_cand_idxs != idx
                men_cand_idxs, men_cand_scores = (
                    men_cand_idxs[filter_mask][:max_knn],
                    men_cand_scores[filter_mask][:max_knn],
                )

                if params["transductive"]:
                    idx += n_train_mentions
                # Add edges to the graphs
                for k in joint_graphs:
                    joint_graph = joint_graphs[k]
                    # Add mention-entity edge
                    joint_graph["rows"] = np.append(
                        joint_graph["rows"], [n_entities + idx]
                    )  # Mentions added at an offset of maximum entities
                    joint_graph["cols"] = np.append(joint_graph["cols"], dict_cand_idx)
                    joint_graph["data"] = np.append(
                        joint_graph["data"], dict_cand_score
                    )
                    if k > 0:
                        # Add mention-mention edges
                        joint_graph["rows"] = np.append(
                            joint_graph["rows"],
                            [n_entities + idx] * len(men_cand_idxs[:k]),
                        )
                        joint_graph["cols"] = np.append(
                            joint_graph["cols"], n_entities + men_cand_idxs[:k]
                        )
                        joint_graph["data"] = np.append(
                            joint_graph["data"], men_cand_scores[:k]
                        )

        if params["transductive"]:
            # Add positive infinity mention-entity edges from training queries to labeled entities
            for idx, train_men in enumerate(train_processed_data):
                dict_cand_idx = train_men["label_idxs"][0]
                for k in joint_graphs:
                    joint_graph = joint_graphs[k]
                    joint_graph["rows"] = np.append(
                        joint_graph["rows"], [n_entities + idx]
                    )  # Mentions added at an offset of maximum entities
                    joint_graph["cols"] = np.append(joint_graph["cols"], dict_cand_idx)
                    joint_graph["data"] = np.append(joint_graph["data"], float("inf"))

        # Compute and print recall metric
        recall_idx_mode = np.argmax(recall_idxs)
        recall_idx_mode_prop = recall_idxs[recall_idx_mode] / np.sum(recall_idxs)
        logger.info(
            f"""
        Recall metrics (for {len(nn_ent_idxs)} queries):
        ---------------"""
        )
        logger.info(
            f"highest recall idx = {recall_idx_mode} ({recall_idxs[recall_idx_mode]}/{np.sum(recall_idxs)} = {recall_idx_mode_prop})"
        )
        for recall_k in recall_accuracy:
            recall_accuracy[recall_k] /= len(nn_ent_idxs)
            logger.info(f"recall@{recall_k} = {recall_accuracy[recall_k]}")

        if params["only_recall"]:
            return

        # Pickle the graphs
        print("Saving joint graphs...")
        with open(graph_path, "wb") as write_handle:
            pickle.dump(joint_graphs, write_handle, protocol=pickle.HIGHEST_PROTOCOL)

        if params["only_embed_and_build"]:
            logger.info(f"Saved embedding data at: {embed_data_path}")
            logger.info(f"Saved graphs at: {graph_path}")
            return

    graph_mode = params.get("graph_mode", None)

    result_overview = {
        "n_entities": n_entities,
        "n_mentions": n_mentions - (n_train_mentions if params["transductive"] else 0),
    }
    results = {}
    if graph_mode is None or graph_mode not in ["directed", "undirected"]:
        results["directed"] = []
        results["undirected"] = []
    else:
        results[graph_mode] = []

    knn_fetch_time = time.time() - time_start
    graph_processing_time = time.time()
    n_graphs_processed = 0.0

    for mode in results:
        print(f"\nEvaluation mode: {mode.upper()}")
        "4) Evaluates the model's accuracy by analyzing how effectively the model can link mentions to the correct entities."
        for k in joint_graphs:
            if k <= knn:
                print(f"\nGraph (k={k}):")
                # Partition graph based on cluster-linking constraints
                partitioned_graph, clusters = eval_cluster_linking.partition_graph(
                    joint_graphs[k],
                    n_entities,
                    mode == "directed",
                    return_clusters=True,
                )
                # Infer predictions from clusters
                result = eval_cluster_linking.analyzeClusters(
                    clusters,
                    entity_data,
                    test_processed_data,
                    k,
                    n_train_mentions if params["transductive"] else 0,
                )
                # log the accuracy
                acc = float(result["accuracy"].split(" ")[0])
                logger.info(f"Eval: accuracy for graph@k={k}: {acc}%")
                # Store result
                results[mode].append(result)
                n_graphs_processed += 1

    avg_graph_processing_time = (
        time.time() - graph_processing_time
    ) / n_graphs_processed
    avg_per_graph_time = (knn_fetch_time + avg_graph_processing_time) / 60

    # Store results
    output_file_name = os.path.join(
        output_path,
        f"eval_results_{__import__('calendar').timegm(__import__('time').gmtime())}",
    )

    try:
        for recall_k in recall_accuracy:
            result_overview[f"recall@{recall_k}"] = recall_accuracy[recall_k]
    except:
        logger.info("Recall data not available since graphs were loaded from disk")

    for mode in results:
        mode_results = results[mode]
        result_overview[mode] = {}
        for r in mode_results:
            k = r["knn_mentions"]
            result_overview[mode][f"accuracy@knn{k}"] = r["accuracy"]
            logger.info(f"{mode} accuracy@knn{k} = {r['accuracy']}")
            output_file = f"{output_file_name}-{mode}-{k}.json"
            with open(output_file, "w") as f:
                json.dump(r, f, indent=2)
                print(f"\nPredictions ({mode}) @knn{k} saved at: {output_file}")
    with open(f"{output_file_name}.json", "w") as f:
        json.dump(result_overview, f, indent=2)
        print(f"\nPredictions overview saved at: {output_file_name}.json")

    # Store results for evaluation object
    base_filename = "biencoder_output_eval"
    biencoder_output_eval = create_versioned_filename(output_path, base_filename)
    with open(f"{biencoder_output_eval}", "w") as f:
        json.dump(output_eval, f, indent=2)
        print(f"\nPredictions overview saved at: {biencoder_output_eval}")

    logger.info(
        "\nThe avg. per graph evaluation time is {} minutes\n".format(
            avg_per_graph_time
        )
    )


def loss_function(
    reranker,
    params,
    forward_output,
    data_module,
    n_entities,
    knn_dict,
    batch_context_inputs,
    accumulate_grad_batches,
):
    """
    Compute the loss function during the training.

    Parameters
    ----------
    - reranker : BiEncoderRanker
        biencoder model
    - params : dict
        Contains most of the relevant keys for training (embed_batch_size, batch_size, n_gpu, force_exact_search etc...)
    - forward_output : dict
        Output of the forward() method in the LitArboel class
    - data_module : Instance of ArboelDataModule class
    - n_entities : int
        Total number of entities
    - knn_dict : int (self.knn_dict = self.hparams["knn"]//2)
        Number of negative entities to fetch. It divides the k-nn evenly between entities and mentions
    - accumulate_grad_batches : int
        Number of steps to accumulate gradients
    """
    device = forward_output["context_inputs"].device
    # Compute the loss
    loss_dual_negs = loss_ent_negs = 0
    # loss of a batch includes both negative mentions and entity inputs (alongside positive examples ofc)
    loss_dual_negs, _ = reranker(
        context_input=forward_output["context_inputs"],
        label_input=forward_output["label_inputs"],
        mst_data={
            "positive_embeds": forward_output["positive_embeds"],
            "negative_dict_inputs": forward_output["negative_dict_inputs"],
            "negative_men_inputs": forward_output["negative_men_inputs"],
        },
        pos_neg_loss=params["pos_neg_loss"],
    )
    skipped_context_inputs = []
    if forward_output["skipped"] > 0 and not params["within_doc_skip_strategy"]:
        # Convert the numpy array to a tensor
        entity_dict_vecs_numpy = [
            data_module.entity_dict_vecs[x].numpy()
            for x in forward_output["skipped_negative_dict_inputs"]
        ]
        single_entity_dict_vecs = np.array(entity_dict_vecs_numpy)
        skipped_negative_dict_inputs = torch.tensor(single_entity_dict_vecs)

        skipped_positive_embeds = []
        for i, pos_idx in enumerate(forward_output["skipped_positive_idxs"]):
            if pos_idx < n_entities:
                data_module.entity_dict_vecs = data_module.entity_dict_vecs.to(device)
                pos_embed = reranker.encode_candidate(
                    cands=data_module.entity_dict_vecs.to(device)[
                        pos_idx : pos_idx + 1
                    ],
                    requires_grad=True,
                )
            else:
                pos_embed = reranker.encode_context(
                    ctxt=data_module.train_men_vecs.to(device)[
                        pos_idx - n_entities : pos_idx - n_entities + 1
                    ],
                    requires_grad=True,
                )
            skipped_positive_embeds.append(pos_embed)
        skipped_positive_embeds = torch.cat(skipped_positive_embeds)
        skipped_context_inputs = batch_context_inputs[
            ~np.array(forward_output["context_inputs_mask"])
        ]
        skipped_label_inputs = torch.tensor(
            [[1] + [0] * (knn_dict)] * len(skipped_context_inputs), dtype=torch.float32
        )
        # Loss of a batch that only includes negative entity inputs.
        loss_ent_negs, _ = reranker(
            context_input=skipped_context_inputs,
            label_input=skipped_label_inputs,
            mst_data={
                "positive_embeds": skipped_positive_embeds,
                "negative_dict_inputs": skipped_negative_dict_inputs,
                "negative_men_inputs": None,
            },
            pos_neg_loss=params["pos_neg_loss"],
        )

    # len(context_input) = Number of mentions in the batch that successfully found negative entities and mentions.
    # len(skipped_context_inputs): Number of mentions in the batch that only found negative entities.
    loss = (
        (
            loss_dual_negs * len(forward_output["context_inputs"])
            + loss_ent_negs * len(skipped_context_inputs)
        )
        / (len(forward_output["context_inputs"]) + len(skipped_context_inputs))
    ) / accumulate_grad_batches

    return loss


"Model Training"


class LitArboel(L.LightningModule):
    def __init__(self, params):
        """
        - params : dict
        Contains most of the relevant keys for training (embed_batch_size, batch_size, force_exact_search etc...)
        """
        super(LitArboel, self).__init__()
        self.save_hyperparameters(params)
        self.reranker = BiEncoderRanker(params)
        self.model = self.reranker.model

        self.best_max_acc = 0
        self.best_embed_and_index_dict = None

    def forward(self, batch_context_inputs, candidate_idxs, n_gold, mention_idxs):
        """
        Description
        -----------
        Processes a batch of input data to generate embeddings, and identifies positive and negative examples for training.
        It handles the construction of mention-entity graphs, computes nearest neighbors, and organizes the data for subsequent loss calculation.

        Parameters
        ----------
        - “batch_context_inputs” : Tensor
            Tensor containing IDs of (mention + surrounding context) tokens. Shape: (batch_size, context_length)
        - “candidate_idxs” : Tensor
            Tensor with indices pointing to the entities in the entity dictionary that are considered correct labels for the mention. Shape: (batch_size, candidate_count)
        - “n_gold” : Tensor
            Number of labels (=entities) associated with the mention. Shape: (batch_size,)
        - “mention_idx” : Tensor
            Tensor containing a sequence of integers from 0 to N-1 (N = number of mentions in the dataset) serbing as a unique identifier for each mention.

        Return
        ------
        - label_inputs : Tensor
            Tensor of binary labels indicating the correct candidates. Shape: (batch_size, 1 + knn_dict + knn_men), where 1 represents the positive example and the rest are negative examples.
        - context_inputs : Tensor
            Processed batch context inputs, filtered to remove mentions with no negative examples. Shape: (filtered_batch_size, context_length).
        - negative_men_inputs : Tensor
            Tensor of negative mention inputs. Shape: (filtered_batch_size * knn_men,)
        - negative_dict_inputs : Tensor
            Tensor of negative dictionary (entity) inputs. Shape: (filtered_batch_size * knn_dict)
        - positive_embeds : Tensor
            Tensor of embeddings for the positive examples. Shape: (filtered_batch_size, embedding_dim)
        - skipped : int
            The number of mentions skipped due to lack of valid negative examples.
        - skipped_positive_idxs : list(int)
            List of indices for positive examples that were skipped.
        - skipped_negative_dict_inputs :
            Tensor of negative dictionary inputs for skipped examples. Shape may vary based on the number of skipped examples and available negative dictionary entries.
        - context_inputs_mask : list(bool)
            Mask indicating which entries in batch_context_inputs were retained after filtering out mentions with no negative examples.
        """

        # mentions within the batch
        mention_embeddings = self.train_men_embeds[mention_idxs.cpu()]
        if len(mention_embeddings.shape) == 1:
            mention_embeddings = np.expand_dims(mention_embeddings, axis=0)
        # Convert Back to Tensor and Move to GPU
        mention_embeddings = torch.from_numpy(mention_embeddings).to(self.device)

        positive_idxs = []
        negative_dict_inputs = []
        negative_men_inputs = []

        skipped_positive_idxs = []
        skipped_negative_dict_inputs = []

        min_neg_mens = float("inf")
        skipped = 0
        context_inputs_mask = [True] * len(batch_context_inputs)

        "IV.4.B) For each mention within the batch"
        # For each mention within the batch
        for m_embed_idx, m_embed in enumerate(mention_embeddings):
            mention_idx = int(mention_idxs[m_embed_idx])
            # ground truth entities of the mention "mention_idx"
            gold_idxs = set(
                self.trainer.datamodule.train_processed_data[mention_idx]["label_idxs"][
                    : n_gold[m_embed_idx]
                ]
            )

            # TEMPORARY: Assuming that there is only 1 gold label, TODO: Incorporate multiple case
            assert n_gold[m_embed_idx] == 1

            if mention_idx in self.gold_links:
                gold_link_idx = self.gold_links[mention_idx]
            else:
                "IV.4.B.a) Create the graph with positive edges"
                # This block creates all the positive edges of the mention in this iteration
                # Run MST on mention clusters of all the gold entities of the current query mention to find its positive edge
                rows, cols, data, shape = (
                    [],
                    [],
                    [],
                    (
                        self.n_entities + self.train_n_mentions,
                        self.n_entities + self.train_n_mentions,
                    ),
                )
                seen = set()

                # Set whether the gold edge should be the nearest or the farthest neighbor
                sim_order = 1 if self.hparams["farthest_neighbor"] else -1

                for cluster_ent in gold_idxs:
                    # IDs of all the mentions inside the gold cluster with entity id = "cluster_ent"
                    cluster_mens = self.trainer.datamodule.train_gold_clusters[
                        cluster_ent
                    ]

                    if self.hparams["within_doc"]:
                        # Filter the gold cluster to within-doc
                        cluster_mens, _ = filter_by_context_doc_id(
                            mention_idxs=cluster_mens,
                            doc_id=self.trainer.datamodule.train_context_doc_ids[
                                mention_idx
                            ],
                            doc_id_list=self.trainer.datamodule.train_context_doc_ids,
                        )

                    # ψ(e, mi) = Enc_E(e)^T Enc_M(mi) for all the mention-entity links inside the cluster of the current mention
                    to_ent_data = (
                        self.train_men_embeds[cluster_mens]
                        @ self.train_dict_embeds[cluster_ent].T
                    )

                    # φ(mi, mj) = Enc_M(mi)^T Enc_M(mj) for all the mention-mention links inside the cluster of the current mention
                    to_men_data = (
                        self.train_men_embeds[cluster_mens]
                        @ self.train_men_embeds[cluster_mens].T
                    )

                    if self.hparams["gold_arbo_knn"] is not None:
                        # Descending order of similarity if nearest-neighbor, else ascending order
                        sorti = np.argsort(sim_order * to_men_data, axis=1)
                        sortv = np.take_along_axis(to_men_data, sorti, axis=1)
                        if self.hparams["rand_gold_arbo"]:
                            randperm = np.random.permutation(sorti.shape[1])
                            sortv, sorti = sortv[:, randperm], sorti[:, randperm]

                    for i in range(len(cluster_mens)):
                        from_node = self.n_entities + cluster_mens[i]
                        to_node = cluster_ent
                        # Add mention-entity link
                        rows.append(from_node)
                        cols.append(to_node)
                        data.append(-1 * to_ent_data[i])  # w_e,mi = - ψ(e, mi)
                        if self.hparams["gold_arbo_knn"] is None:
                            # Add forward and reverse mention-mention links over the entire MST
                            for j in range(i + 1, len(cluster_mens)):
                                to_node = self.n_entities + cluster_mens[j]
                                if (from_node, to_node) not in seen:
                                    score = to_men_data[i, j]
                                    rows.append(from_node)
                                    cols.append(to_node)
                                    # w_i,j = -ψ(mi, mj)
                                    data.append(
                                        -1 * score
                                    )  # Negatives needed for SciPy's Minimum Spanning Tree computation
                                    seen.add((from_node, to_node))
                                    seen.add((to_node, from_node))
                        else:
                            # Approximate the MST using <gold_arbo_knn> nearest mentions from the gold cluster
                            added = 0
                            approx_k = min(
                                self.hparams["gold_arbo_knn"] + 1, len(cluster_mens)
                            )
                            for j in range(approx_k):
                                if added == approx_k - 1:
                                    break
                                to_node = self.n_entities + cluster_mens[sorti[i, j]]
                                if to_node == from_node:
                                    continue
                                added += 1
                                if (from_node, to_node) not in seen:
                                    score = sortv[i, j]
                                    rows.append(from_node)
                                    cols.append(to_node)
                                    data.append(
                                        -1 * score
                                    )  # Negatives needed for SciPy's Minimum Spanning Tree computation
                                    seen.add((from_node, to_node))

                "IV.4.B.b) Fine tuning with inference procedure to get a mst"
                # Creates MST with entity constraint (inference procedure)
                csr = csr_matrix(
                    (-sim_order * np.array(data), (rows, cols)), shape=shape
                )
                # Note: minimum_spanning_tree expects distances as edge weights
                mst = minimum_spanning_tree(csr).tocoo()
                # Note: cluster_linking_partition expects similarities as edge weights # Convert directed to undirected graph
                rows, cols, data = cluster_linking_partition(
                    np.concatenate(
                        (mst.row, mst.col)
                    ),  # cluster_linking_partition is imported from eval_cluster_linking
                    np.concatenate((mst.col, mst.row)),
                    np.concatenate((sim_order * mst.data, sim_order * mst.data)),
                    self.n_entities,
                    directed=True,
                    silent=True,
                )
                assert np.array_equal(rows - self.n_entities, cluster_mens)

                for i in range(len(rows)):
                    men_idx = rows[i] - self.n_entities
                    if men_idx in self.gold_links:
                        continue
                    assert men_idx >= 0
                    add_link = True
                    # Store the computed positive edges for the mentions in the clusters only if they have the same gold entities as the query mention
                    for l in self.trainer.datamodule.train_processed_data[men_idx][
                        "label_idxs"
                    ][
                        : self.trainer.datamodule.train_processed_data[men_idx][
                            "n_labels"
                        ]
                    ]:
                        if l not in gold_idxs:
                            add_link = False
                            break
                    if add_link:
                        self.gold_links[men_idx] = cols[i]
                gold_link_idx = self.gold_links[mention_idx]

            "IV.4.B.c) Retrieve the pre-computed nearest neighbors"
            knn_dict_idxs = self.dict_nns[mention_idx]
            knn_dict_idxs = knn_dict_idxs.astype(np.int64).flatten()
            knn_men_idxs = self.men_nns[mention_idx][self.men_nns[mention_idx] != -1]
            knn_men_idxs = knn_men_idxs.astype(np.int64).flatten()
            if self.hparams["within_doc"]:
                knn_men_idxs, _ = filter_by_context_doc_id(
                    mention_idxs=knn_men_idxs,
                    doc_id=self.trainer.datamodule.train_context_doc_ids[mention_idx],
                    doc_id_list=self.trainer.datamodule.train_context_doc_ids,
                    return_numpy=True,
                )
            "IV.4.B.d) Add negative examples"
            neg_mens = list(
                knn_men_idxs[
                    ~np.isin(
                        knn_men_idxs,
                        np.concatenate(
                            [
                                self.trainer.datamodule.train_gold_clusters[gi]
                                for gi in gold_idxs
                            ]
                        ),
                    )
                ][: self.knn_men]
            )
            # Track queries with no valid mention negatives
            if len(neg_mens) == 0:
                context_inputs_mask[m_embed_idx] = False
                skipped_negative_dict_inputs += list(
                    knn_dict_idxs[~np.isin(knn_dict_idxs, list(gold_idxs))][
                        : self.knn_dict
                    ]
                )
                skipped_positive_idxs.append(gold_link_idx)
                skipped += 1
                continue
            else:
                min_neg_mens = min(min_neg_mens, len(neg_mens))
            negative_men_inputs.append(
                knn_men_idxs[
                    ~np.isin(
                        knn_men_idxs,
                        np.concatenate(
                            [
                                self.trainer.datamodule.train_gold_clusters[gi]
                                for gi in gold_idxs
                            ]
                        ),
                    )
                ][: self.knn_men]
            )
            negative_dict_inputs += list(
                knn_dict_idxs[~np.isin(knn_dict_idxs, list(gold_idxs))][: self.knn_dict]
            )
            # Add the positive example
            positive_idxs.append(gold_link_idx)

        "IV.4.C) Skip this iteration if no suitable negative examples found"
        if len(negative_men_inputs) == 0:
            return None

        # Sets the minimum number of negative mentions found across all processed mentions in the current batch
        self.knn_men = min_neg_mens

        # This step ensures that each mention is compared against a uniform number of negative mentions
        filtered_negative_men_inputs = []
        for row in negative_men_inputs:
            filtered_negative_men_inputs += list(row[: self.knn_men])
        negative_men_inputs = filtered_negative_men_inputs
        # Assertions for Data Integrity
        assert (
            len(negative_dict_inputs)
            == (len(mention_embeddings) - skipped) * self.knn_dict
        )
        assert (
            len(negative_men_inputs)
            == (len(mention_embeddings) - skipped) * self.knn_men
        )

        self.total_skipped += skipped
        self.total_knn_men_negs += self.knn_men

        entity_dict_vecs_numpy = [
            self.trainer.datamodule.entity_dict_vecs[x].numpy()
            for x in negative_dict_inputs
        ]
        # Then, convert the list of numpy arrays into a single numpy array
        single_entity_dict_vecs = np.array(entity_dict_vecs_numpy)
        # Finally, convert the numpy array to a tensor
        negative_dict_inputs = torch.tensor(single_entity_dict_vecs)

        train_men_vecs_numpy = [
            self.trainer.datamodule.train_men_vecs[x].numpy()
            for x in negative_men_inputs
        ]
        # Then, convert the list of numpy arrays into a single numpy array
        single_train_men_vecs = np.array(train_men_vecs_numpy)
        # Finally, convert the numpy array to a tensor
        negative_men_inputs = torch.tensor(single_train_men_vecs)

        # Labels indicating the correct candidates. Used for computing loss.
        positive_embeds = []
        for i, pos_idx in enumerate(positive_idxs):
            if pos_idx < self.n_entities:
                pos_embed = self.reranker.encode_candidate(
                    cands=self.trainer.datamodule.entity_dict_vecs.to(self.device)[
                        pos_idx : pos_idx + 1
                    ],
                    requires_grad=True,
                )
            else:
                pos_embed = self.reranker.encode_context(
                    ctxt=self.trainer.datamodule.train_men_vecs.to(self.device)[
                        pos_idx - self.n_entities : pos_idx - self.n_entities + 1
                    ],
                    requires_grad=True,
                )
            positive_embeds.append(pos_embed)
        positive_embeds = torch.cat(positive_embeds)

        # Remove mentions with no negative examples
        context_inputs = batch_context_inputs[context_inputs_mask]

        # Tensor containing binary values that act as indicator variables in the paper:
        # Contains Indicator variable such that I_{u,m_i} = 1 if(u,mi) ∈ E'_{m_i} and I{u,m_i} = 0 otherwise.
        label_inputs = torch.tensor(
            [[1] + [0] * (self.knn_dict + self.knn_men)] * len(context_inputs),
            dtype=torch.float32,
        ).to(self.device)

        return {
            "label_inputs": label_inputs,
            "context_inputs": context_inputs,
            "negative_men_inputs": negative_men_inputs,
            "negative_dict_inputs": negative_dict_inputs,
            "positive_embeds": positive_embeds,
            "skipped": skipped,
            "skipped_positive_idxs": skipped_positive_idxs,
            "skipped_negative_dict_inputs": skipped_negative_dict_inputs,
            "context_inputs_mask": context_inputs_mask,
        }

    def training_step(self, batch, batch_idx):
        # batch is a subsample from tensor_dataset
        batch_context_inputs, candidate_idxs, n_gold, mention_idxs = batch

        f = self.forward(
            batch_context_inputs=batch_context_inputs,
            candidate_idxs=candidate_idxs,
            n_gold=n_gold,
            mention_idxs=mention_idxs,
        )

        for key, value in f.items():
            if isinstance(value, torch.Tensor):
                f[key] = value.to(self.device)

        # Compute the loss
        train_loss = loss_function(
            reranker=self.reranker,
            params=self.hparams,
            forward_output=f,
            data_module=self.trainer.datamodule,
            n_entities=self.n_entities,
            knn_dict=self.knn_dict,
            batch_context_inputs=batch_context_inputs,
            accumulate_grad_batches=self.trainer.accumulate_grad_batches,
        )
        print("training loss :", train_loss)
        self.log(
            "train loss :", train_loss, on_step=True, on_epoch=True, sync_dist=True
        )

        return train_loss

    def configure_optimizers(self):

        # Define optimizer
        optimizer = get_bert_optimizer(
            models=[self.model],
            type_optimization=self.hparams["type_optimization"],
            learning_rate=self.hparams["learning_rate"],
            fp16=self.hparams.get("fp16"),
        )

        # Define scheduler
        num_train_steps = (
            int(
                len(self.trainer.datamodule.train_tensor_data)
                / self.hparams["train_batch_size"]
                / self.trainer.accumulate_grad_batches
            )
            * self.trainer.max_epochs
        )
        num_warmup_steps = int(num_train_steps * self.hparams["warmup_proportion"])

        scheduler = WarmupLinearSchedule(
            optimizer=optimizer,
            warmup_steps=num_warmup_steps,
            t_total=num_train_steps,
        )
        logger.info(" Num optimization steps = %d" % num_train_steps)
        logger.info(" Num warmup steps = %d", num_warmup_steps)
        return [optimizer], [{"scheduler": scheduler, "interval": "step"}]

    def on_train_epoch_start(self):

        logger.info("On train epoch start")
        "IV.1) Compute mention and entity embeddings and indexes at the start of each epoch"
        # Compute mention and entity embeddings and indexes at the start of each epoch
        # Corpus is a collection of entities, which is used to build type-specific search indexes if provided.
        """
        With a Corpus : Multiple type-specific indexes are created, allowing for more targeted and efficient searches within specific categories of entities.
        'dict_embeds' and 'men_embeds': The resulting entity and mention embeddings.
        'dict_indexes' and 'men_indexes': Dictionary that will store search indexes (!= indices)for each unique entity type found in the corpus
        'dict_idxs_by_type' and 'men_idxs_by_type': Dictionary to store indices of the corpus elements, grouped by their entity type.
        !!! idxs = indices / indexes = indexes !!!
        """
        logger.info(
            "TRAINING. Dictionary: Embedding and building index"
        )  # For entities
        if self.hparams["use_types"]:  # type-specific indexes
            (
                self.train_dict_embeds,
                self.train_dict_indexes,
                self.train_dict_idxs_by_type,
            ) = data_process.embed_and_index(
                model=self.reranker,
                token_id_vecs=self.trainer.datamodule.entity_dict_vecs,
                encoder_type="candidate",
                corpus=self.trainer.datamodule.entity_dictionary,
                force_exact_search=self.hparams["force_exact_search"],
                batch_size=self.hparams["embed_batch_size"],
                probe_mult_factor=self.hparams["probe_mult_factor"],
            )
            (
                self.train_men_embeds,
                self.train_men_indexes,
                self.train_men_idxs_by_type,
            ) = data_process.embed_and_index(
                model=self.reranker,
                token_id_vecs=self.trainer.datamodule.train_men_vecs,
                encoder_type="context",
                corpus=self.trainer.datamodule.train_processed_data,
                force_exact_search=self.hparams["force_exact_search"],
                batch_size=self.hparams["embed_batch_size"],
                probe_mult_factor=self.hparams["probe_mult_factor"],
            )

        else:  # general indexes
            self.train_dict_embeds, self.train_dict_index = (
                data_process.embed_and_index(
                    model=self.reranker,
                    token_id_vecs=self.trainer.datamodule.entity_dict_vecs,
                    encoder_type="candidate",
                    force_exact_search=self.hparams["force_exact_search"],
                    batch_size=self.hparams["embed_batch_size"],
                    probe_mult_factor=self.hparams["probe_mult_factor"],
                )
            )
            self.train_men_embeds, self.train_men_index = data_process.embed_and_index(
                model=self.reranker,
                token_id_vecs=self.trainer.datamodule.train_men_vecs,
                encoder_type="context",
                force_exact_search=self.hparams["force_exact_search"],
                batch_size=self.hparams["embed_batch_size"],
                probe_mult_factor=self.hparams["probe_mult_factor"],
            )

        # Number of entities
        self.n_entities = len(self.trainer.datamodule.entity_dictionary)
        # Number of mentions in training set
        self.train_n_mentions = len(self.trainer.datamodule.train_processed_data)

        # Store golden MST links
        self.gold_links = {}
        # Calculate the number of negative entities and mentions to fetch # Divides the k-nn evenly between entities and mentions
        self.knn_dict = self.hparams["knn"] // 2
        self.knn_men = self.hparams["knn"] - self.knn_dict

        "3) knn search : indice and distance of k closest mentions and entities"
        logger.info("TRAINING. Starting KNN search...")
        # INFO: Fetching all sorted mentions to be able to filter to within-doc later=
        n_men_to_fetch = (
            len(self.train_men_embeds)
            if self.hparams["use_types"]
            else self.knn_men + self.trainer.datamodule.max_gold_cluster_len
        )
        n_ent_to_fetch = (
            self.knn_dict + 1
        )  # +1 accounts for the possibility of self-reference
        if not self.hparams["use_types"]:
            _, self.dict_nns = self.train_dict_index.search(
                self.train_men_embeds, n_ent_to_fetch
            )
            _, self.men_nns = self.train_men_index.search(
                self.train_men_embeds, n_men_to_fetch
            )
        else:
            self.dict_nns = -1 * np.ones((len(self.train_men_embeds), n_ent_to_fetch))
            self.men_nns = -1 * np.ones((len(self.train_men_embeds), n_men_to_fetch))
            for entity_type in self.train_men_indexes:
                self.men_embeds_by_type = self.train_men_embeds[
                    self.train_men_idxs_by_type[entity_type]
                ]
                _, self.dict_nns_by_type = self.train_dict_indexes[entity_type].search(
                    self.men_embeds_by_type, n_ent_to_fetch
                )
                _, self.men_nns_by_type = self.train_men_indexes[entity_type].search(
                    self.men_embeds_by_type,
                    min(n_men_to_fetch, len(self.men_embeds_by_type)),
                )
                self.dict_nns_idxs = np.array(
                    list(
                        map(
                            lambda x: self.train_dict_idxs_by_type[entity_type][x],
                            self.dict_nns_by_type,
                        )
                    )
                )
                self.men_nns_idxs = np.array(
                    list(
                        map(
                            lambda x: self.train_men_idxs_by_type[entity_type][x],
                            self.men_nns_by_type,
                        )
                    )
                )
                for i, idx in enumerate(self.train_men_idxs_by_type[entity_type]):
                    self.dict_nns[idx] = self.dict_nns_idxs[i]
                    self.men_nns[idx][: len(self.men_nns_idxs[i])] = self.men_nns_idxs[
                        i
                    ]
        logger.info("TRAINING. Search finished")

        self.total_skipped = self.total_knn_men_negs = 0

    def on_validation_epoch_start(self):

        torch.cuda.empty_cache()  # Empty the CUDA cache to free up GPU memory

        self.valid_n_mentions = len(
            self.trainer.datamodule.valid_processed_data
        )  # total number of mentions in validation set
        self.valid_max_knn = 8  # max number of neighbors

        "Computes embeddings and indexes for entities and mentions. "
        "This block is preparing the data for evaluation by transforming raw vectors into a format that can be efficiently used for retrieval and comparison operations"
        if self.hparams["use_types"]:  # corpus = entity data
            # corpus is a collection of entities, which is used to build type-specific search indexes if provided.
            """
            With a Corpus : Multiple type-specific indexes are created, allowing for more targeted and efficient searches within specific categories of entities.
            'dict_embeds' and 'men_embeds': The resulting entity and mention embeddings.
            'dict_indexes' and 'men_indexes': Dictionary that will store search indexes (!= indices)for each unique entity type found in the corpus
            'dict_idxs_by_type' and 'men_idxs_by_type': Dictionary to store indices of the corpus elements, grouped by their entity type.
            !!! idxs = indices / indexes = indexes !!!
            """
            logger.info(
                "VALIDATION. Dictionary: Embedding and building index"
            )  # For entities
            (
                self.valid_dict_embeds,
                self.valid_dict_indexes,
                self.valid_dict_idxs_by_type,
            ) = data_process.embed_and_index(
                model=self.reranker,
                token_id_vecs=self.trainer.datamodule.entity_dict_vecs,
                encoder_type="candidate",
                corpus=self.trainer.datamodule.entity_dictionary,
                force_exact_search=self.hparams["force_exact_search"],
                batch_size=self.hparams["embed_batch_size"],
                probe_mult_factor=self.hparams["probe_mult_factor"],
            )
            logger.info(
                "VALIDATION. Queries: Embedding and building index"
            )  # For mentions
            (
                self.valid_men_embeds,
                self.valid_men_indexes,
                self.valid_men_idxs_by_type,
            ) = data_process.embed_and_index(
                model=self.reranker,
                token_id_vecs=self.trainer.datamodule.valid_men_vecs,
                encoder_type="context",
                corpus=self.trainer.datamodule.valid_processed_data,
                force_exact_search=self.hparams["force_exact_search"],
                batch_size=self.hparams["embed_batch_size"],
                probe_mult_factor=self.hparams["probe_mult_factor"],
            )
        else:  # corpus = None
            """
            Without a Corpus: A single, general index is created for all embeddings, suitable for broad searches across the entire dataset.
            'dict_embeds' and 'men_embeds': The resulting entity and mention embeddings.
            'dict_index' and 'men_index': Dictionary that will store search index
            """
            logger.info("VALIDATION. Dictionary: Embedding and building index")
            self.valid_dict_embeds, self.valid_dict_index = (
                data_process.embed_and_index(
                    model=self.reranker,
                    token_id_vecs=self.trainer.datamodule.entity_dict_vecs,
                    encoder_type="candidate",
                    force_exact_search=self.hparams["force_exact_search"],
                    batch_size=self.hparams["embed_batch_size"],
                    probe_mult_factor=self.hparams["probe_mult_factor"],
                )
            )
            logger.info("VALIDATION. Queries: Embedding and building index")
            self.valid_men_embeds, self.valid_men_index = data_process.embed_and_index(
                model=self.reranker,
                token_id_vecs=self.trainer.datamodule.valid_men_vecs,
                encoder_type="context",
                force_exact_search=self.hparams["force_exact_search"],
                batch_size=self.hparams["embed_batch_size"],
                probe_mult_factor=self.hparams["probe_mult_factor"],
            )

        "Performs k-nearest neighbors (k-NN) search to establish relationships between mentions and entities."
        logger.info("VALIDATION. Eval: Starting KNN search...")
        # Fetch (k+1) NN mention candidates; fetching all mentions for within_doc to filter down later
        self.valid_n_men_to_fetch = (
            len(self.valid_men_embeds)
            if self.hparams["within_doc"]
            else self.valid_max_knn + 1
        )  # Number of mentions to fetch
        if not self.hparams["use_types"]:  # Only one index so only need one search
            self.valid_nn_ent_dists, self.valid_nn_ent_idxs = (
                self.valid_dict_index.search(self.valid_men_embeds, 1)
            )  # Returns the distance and the indice of the closest entity for all mentions in "men_embeds"
            self.valid_nn_men_dists, self.valid_nn_men_idxs = (
                self.valid_men_index.search(
                    self.valid_men_embeds, self.valid_n_men_to_fetch
                )
            )  # Returns the distances and the indices of the k closest mentions for all mentions in "men_embeds"
        else:  # Several indexes corresponding to the different entities in entity_data so we can use the specific search index
            self.valid_nn_ent_idxs = -1 * np.ones(
                (len(self.valid_men_embeds), 1), dtype=int
            )  # Indice of the closest entity for all mentions in "men_embeds"
            self.valid_nn_ent_dists = -1 * np.ones(
                (len(self.valid_men_embeds), 1), dtype="float64"
            )  # Distance of the closest entity for all mentions in "men_embeds"
            self.valid_nn_men_idxs = -1 * np.ones(
                (len(self.valid_men_embeds), self.valid_n_men_to_fetch), dtype=int
            )  # Indice of k closest mentions for all mentions in "men_embeds"
            self.valid_nn_men_dists = -1 * np.ones(
                (len(self.valid_men_embeds), self.valid_n_men_to_fetch), dtype="float64"
            )  # Distance of the k closest mentions for all mentions in "men_embeds"
            for entity_type in self.valid_men_indexes:
                # Creates a new list only containing the mentions for which type = entity_types
                self.valid_men_embeds_by_type = self.valid_men_embeds[
                    self.valid_men_idxs_by_type[entity_type]
                ]  # Only want to search the mentions that belongs to a specific type of entity.
                # Returns the distance and the indice of the closest entity for all mentions in "men_embeds" by entity type
                self.valid_nn_ent_dists_by_type, nn_ent_idxs_by_type = (
                    self.valid_dict_indexes[entity_type].search(
                        self.valid_men_embeds_by_type, 1
                    )
                )
                self.valid_nn_ent_idxs_by_type = np.array(
                    list(
                        map(
                            lambda x: self.valid_dict_idxs_by_type[entity_type][x],
                            nn_ent_idxs_by_type,
                        )
                    )
                )
                # Returns the distance and the indice of the k closest mentions for all mention in "men_embeds" by entity type
                # Note that here we may not necessarily have k mentions in each entity type which is why we use min(k,len(self.valid_men_embeds_by_type))
                (
                    self.valid_nn_men_dists_by_type,
                    nn_men_idxs_by_type,
                ) = self.valid_men_indexes[entity_type].search(
                    self.valid_men_embeds_by_type,
                    min(self.valid_n_men_to_fetch, len(self.valid_men_embeds_by_type)),
                )
                self.valid_nn_men_idxs_by_type = np.array(
                    list(
                        map(
                            lambda x: self.valid_men_idxs_by_type[entity_type][x],
                            nn_men_idxs_by_type,
                        )
                    )
                )
                for i, idx in enumerate(self.valid_men_idxs_by_type[entity_type]):
                    self.valid_nn_ent_idxs[idx] = self.valid_nn_ent_idxs_by_type[i]
                    self.valid_nn_ent_dists[idx] = self.valid_nn_ent_dists_by_type[i]
                    self.valid_nn_men_idxs[idx][
                        : len(self.valid_nn_men_idxs_by_type[i])
                    ] = self.valid_nn_men_idxs_by_type[i]
                    self.valid_nn_men_dists[idx][
                        : len(self.valid_nn_men_dists_by_type[i])
                    ] = self.valid_nn_men_dists_by_type[i]
        logger.info("VALIDATION. Eval: Search finished")

        self.embed_and_index_dict = (
            {
                "dict_embeds": self.valid_dict_embeds,
                "dict_indexes": self.valid_dict_indexes,
                "dict_idxs_by_type": self.valid_dict_idxs_by_type,
            }
            if self.hparams["use_types"]
            else {
                "dict_embeds": self.valid_dict_embeds,
                "dict_index": self.valid_dict_index,
            }
        )

        pass

    def validation_step(self, batch_, batch_idx):

        self.n_entities = len(self.trainer.datamodule.entity_dictionary)
        self.max_acc, self.dict_acc = evaluate(
            params=self.hparams,
            logger=logger,
            entity_data=self.trainer.datamodule.entity_dictionary,
            query_data=self.trainer.datamodule.valid_processed_data,
            valid_men_embeds=self.valid_men_embeds,
            nn_ent_dists=self.valid_nn_ent_dists,
            nn_ent_idxs=self.valid_nn_ent_idxs,
            nn_men_dists=self.valid_nn_men_dists,
            nn_men_idxs=self.valid_nn_men_idxs,
            n_entities=self.n_entities,
            n_mentions=self.valid_n_mentions,
            max_knn=self.valid_max_knn,
            use_types=self.hparams["use_types"],
            within_doc=self.hparams["within_doc"],
            context_doc_ids=self.trainer.datamodule.valid_context_doc_ids,
        )

    def on_validation_epoch_end(self):
        self.log("max_acc", self.max_acc, on_epoch=True, sync_dist=True)

        for key, value in self.dict_acc.items():
            self.log(f"dict_acc_{key}", value, on_epoch=True, sync_dist=True)

        if self.max_acc > self.best_max_acc:
            self.best_max_acc = self.max_acc
            self.best_embed_and_index_dict = self.embed_and_index_dict
            save_path = os.path.join(
                self.hparams.get("output_path"),
                f"embed_and_index_dict_:epoch_{self.current_epoch}.pth",
            )
            torch.save(self.best_embed_and_index_dict, save_path, pickle_protocol=4)

    def test_step(self, batch_, batch_idx):

        evaluate_test(
            params=self.hparams,
            reranker=self.reranker,
            test_dict_vecs=self.trainer.datamodule.entity_dict_vecs,
            test_processed_data=self.trainer.datamodule.test_processed_data,
            test_men_vecs=self.trainer.datamodule.test_men_vecs,
            logger=logger,
            entity_data=self.trainer.datamodule.entity_dictionary,
            train_processed_data=(
                self.trainer.datamodule.train_processed_data
                if self.hparams.get("transductive")
                else None
            ),
            train_men_vecs=(
                self.trainer.datamodule.train_men_vecs
                if self.hparams.get("transductive")
                else None
            ),
            use_types=self.hparams["use_types"],
            embed_batch_size=self.hparams["embed_batch_size"],
            force_exact_search=self.hparams["force_exact_search"],
            probe_mult_factor=self.hparams["probe_mult_factor"],
            within_doc=self.hparams["within_doc"],
            context_doc_ids=(
                self.trainer.datamodule.test_context_doc_ids
                if self.hparams["within_doc"]
                else None
            ),
        )

    def on_train_start(self):
        self.start_time = time.time()

    def on_train_end(self):
        if self.start_time is not None:
            execution_time = (time.time() - self.start_time) / 60
            logging.info(f"The training took {execution_time:.2f} minutes")
        else:
            logging.warning("Start time not set. Unable to calculate execution time.")

    def on_test_start(self):
        self.start_time = time.time()

    def on_test_end(self):
        if self.start_time is not None:
            execution_time = (time.time() - self.start_time) / 60
            logging.info(f"The testing took {execution_time:.2f} minutes")
        else:
            logging.warning("Start time not set. Unable to calculate execution time.")



import unittest
import numpy as np

from tqdm import tqdm

from bioel.ontology import BiomedicalOntology
from bioel.model import Model_Wrapper
import os
import pickle
import ujson


class TestModel(unittest.TestCase):

    def test_arboel_biencoder(self):
        """
        TestCase - 1: Test the training part of Biencoder
        """
        with open("config_arboel.json", "r") as f:
            config = ujson.load(f)
        test_cases = config["test_cases_biencoder"]
        for params_file in test_cases:
            arboel_biencoder = Model_Wrapper.load_arboel_biencoder(
                name="arboel_biencoder",
                params_file=params_file,
            )
            arboel_biencoder.training()
            arboel_biencoder.inference()

    def test_arboel_crossencoder(self):
        """
        TestCase - 1: Test the training part of Biencoder
        """
        with open("config_arboel.json", "r") as f:
            config = ujson.load(f)
        test_cases = config["test_cases_crossencoder"]
        for params_file in test_cases:
            arboel_cross = Model_Wrapper.load_arboel_crossencoder(
                name="arboel_crossencoder",
                params_file=params_file,
            )
            arboel_cross.training()
            arboel_cross.inference()


# class TestModel(unittest.TestCase):

#     def test_train_arboel_biencoder(self):
#         """
#         TestCase - 1: Test the training part of Biencoder
#         """

#         # Load configuration file
#         with open("config_arboel.json", "r") as f:
#             config = ujson.load(f)

#         test_cases = config["test_cases_train_biencoder"]

#         parser = BlinkParser(add_model_args=True)
#         parser.add_training_args()

#         for case in test_cases:
#             test_args = [
#                 "--data_path",
#                 case["data_path"],
#                 "--output_path",
#                 case["output_path"],
#                 "--dataset",
#                 case["dataset"],
#                 "--ontology",
#                 case["ontology"],
#                 "--load_function",
#                 case["load_function"],
#                 "--model_name_or_path",
#                 "dmis-lab/biobert-base-cased-v1.2",
#             ]

#             args = parser.parse_args(test_args)
#             args_dict = args.__dict__
#             args_dict["ontology_dict"] = case["ontology_dict"]

#             # Check if the function name provided is an actual function of BiomedicalOntology
#             if hasattr(BiomedicalOntology, args_dict["load_function"]):
#                 load_func = getattr(BiomedicalOntology, args_dict["load_function"])

#                 if args_dict["ontology_dict"]:
#                     ontology_object = load_func(**args_dict["ontology_dict"])
#                     print(f"Ontology loaded successfully. Name: {ontology_object.name}")
#                 else:
#                     raise ValueError("No ontology data provided.")
#             else:
#                 raise ValueError(
#                     f"Error: {args_dict['load_function']} is not a valid function for BiomedicalOntology."
#                 )

#             with open(
#                 os.path.join(
#                     args_dict["data_path"], f"{args_dict['ontology']}_object.pickle"
#                 ),
#                 "wb",
#             ) as f:
#                 pickle.dump(ontology_object, f)

#             data_module = ArboelDataModule(params=args_dict)

#             model = LitArboel(params=args_dict)

#             trainer = L.Trainer(
#                 num_sanity_val_steps=0,
#                 limit_train_batches=1,
#                 limit_val_batches=1,
#                 max_epochs=1,
#                 devices=[0],
#                 accelerator="gpu",
#                 strategy="ddp_find_unused_parameters_true",
#             )

#             trainer.fit(model, datamodule=data_module)

#     def test_evaluate_arboel_biencoder(self):
#         """
#         TestCase - 2: Test the evaluate part of Biencoder
#         """

#         # Load configuration file
#         with open("config_arboel.json", "r") as f:
#             config = ujson.load(f)

#         test_cases = config["test_cases_eval_biencoder"]

#         parser = BlinkParser(add_model_args=True)
#         parser.add_eval_args()

#         for case in test_cases:

#             test_args = [
#                 "--data_path",
#                 case["data_path"],
#                 "--output_path",
#                 case["output_path"],
#                 "--dataset",
#                 case["dataset"],
#                 "--ontology",
#                 case["ontology"],
#                 "--load_function",
#                 case["load_function"],
#                 "--biencoder_checkpoint",
#                 case["biencoder_checkpoint"],
#                 "--model_name_or_path",
#                 "dmis-lab/biobert-base-cased-v1.2",
#             ]

#             args = parser.parse_args(test_args)
#             args_dict = args.__dict__
#             args_dict["ontology_dict"] = case["ontology_dict"]
#             # Check if the function name provided is an actual function of BiomedicalOntology
#             if hasattr(BiomedicalOntology, args_dict["load_function"]):
#                 load_func = getattr(BiomedicalOntology, args_dict["load_function"])

#                 if args_dict["ontology_dict"]:
#                     ontology_object = load_func(**args_dict["ontology_dict"])
#                     print(f"Ontology loaded successfully. Name: {ontology_object.name}")
#                 else:
#                     raise ValueError("No ontology data provided.")
#             else:
#                 raise ValueError(
#                     f"Error: {args_dict['load_function']} is not a valid function for BiomedicalOntology."
#                 )

#             with open(
#                 os.path.join(
#                     args_dict["data_path"], f"{args_dict['ontology']}_object.pickle"
#                 ),
#                 "wb",
#             ) as f:
#                 pickle.dump(ontology_object, f)

#             data_module = ArboelDataModule(params=args_dict)

#             model = LitArboel.load_from_checkpoint(
#                 params=args_dict, checkpoint_path=args_dict["biencoder_checkpoint"]
#             )

#             trainer = L.Trainer(
#                 limit_test_batches=1,
#                 devices=[0],
#                 accelerator="gpu",
#                 strategy="ddp_find_unused_parameters_true",
#             )

#             trainer.test(model=model, datamodule=data_module)

#     def test_train_arboel_crossencoder(self):
#         """
#         TestCase - 3: Test the training part of Crossencoder
#         """

#         # Load configuration file
#         with open("config_arboel.json", "r") as f:
#             config = ujson.load(f)

#         test_cases = config["test_cases_train_crossencoder"]

#         parser = BlinkParser(add_model_args=True)
#         parser.add_training_args()

#         for case in test_cases:
#             test_args = [
#                 "--data_path",
#                 case["data_path"],
#                 "--output_path",
#                 case["output_path"],
#                 "--biencoder_checkpoint",
#                 case["biencoder_checkpoint"],
#                 "--biencoder_indices_path",
#                 case["biencoder_indices_path"],
#                 "--model_name_or_path",
#                 "dmis-lab/biobert-base-cased-v1.2",
#                 "--add_linear",
#                 "--train_batch_size",
#                 "2",
#             ]

#             args = parser.parse_args(test_args)
#             args_dict = args.__dict__

#             data_module = CrossEncoderDataModule(params=args_dict)

#             model = LitCrossEncoder(params=args_dict)

#             trainer = L.Trainer(
#                 limit_train_batches=1,
#                 limit_val_batches=1,
#                 num_sanity_val_steps=0,
#                 max_epochs=1,
#                 devices=[0],
#                 accelerator="gpu",
#                 strategy="ddp_find_unused_parameters_true",
#             )

#             trainer.fit(model, datamodule=data_module)

#     def test_evaluate_arboel_crossencoder(self):
#         """
#         TestCase - 4: Test the evaluate part of Crossencoder
#         """

#         # Load configuration file
#         with open("config_arboel.json", "r") as f:
#             config = ujson.load(f)

#         test_cases = config["test_cases_eval_crossencoder"]

#         parser = BlinkParser(add_model_args=True)
#         parser.add_eval_args()
#         for case in test_cases:

#             test_args = [
#                 "--data_path",
#                 case["data_path"],
#                 "--output_path",
#                 case["output_path"],
#                 "--biencoder_checkpoint",
#                 case["biencoder_checkpoint"],
#                 "--biencoder_indices_path",
#                 case["biencoder_indices_path"],
#                 "--crossencoder_checkpoint",
#                 case["crossencoder_checkpoint"],
#                 "--model_name_or_path",
#                 "dmis-lab/biobert-base-cased-v1.2",
#                 "--add_linear",
#             ]

#             args = parser.parse_args(test_args)
#             args_dict = args.__dict__
#             args_dict["recall_k_list"] = [2]

#             data_module = CrossEncoderDataModule(params=args_dict)

#             model = LitCrossEncoder.load_from_checkpoint(
#                 params=args_dict, checkpoint_path=args_dict["crossencoder_checkpoint"]
#             )

#             trainer = L.Trainer(
#                 limit_test_batches=1,
#                 devices=[0],
#                 accelerator="gpu",
#                 strategy="ddp_find_unused_parameters_true",
#             )

#             trainer.test(model=model, datamodule=data_module)



from bioel.ontology import BiomedicalOntology, BiomedicalEntity
from typing import List, Optional, Union, Dict, NamedTuple, Set, Tuple, Type
from bioel.utils.file_cache import get_path
from collections import defaultdict
import json
import datetime
import numpy
import scipy
import joblib
from sklearn.feature_extraction.text import TfidfVectorizer
import nmslib
from nmslib.dist import FloatIndex
import re


class LinkerPaths(NamedTuple):
    """
    Encapsulates all the (possibly remote) paths to data for a scispacy CandidateGenerator.
    ann_index: str
        Path to the approximate nearest neighbours index.
    tfidf_vectorizer: str
        Path to the joblib serialized sklearn TfidfVectorizer.
    tfidf_vectors: str
        Path to the float-16 encoded tf-idf vectors for the entities in the KB.
    concept_aliases_list: str
        Path to the indices mapping concepts to aliases in the index.
    alias_to_cuis: Dict[str, Set[str]]
        Dictionary which maps aliases to their unique concept ids
    cui_to_entity : Dict[str, BiomedicalEntity]
        Dictionary which maps unique concept ids to the corresponding entities
    """

    ann_index: str
    tfidf_vectorizer: str
    tfidf_vectors: str
    concept_aliases_list: str
    alias_to_cuis: Dict[str, Set[str]]
    cui_to_entity: Dict[str, BiomedicalEntity]


class KnowledgeBaseEmbeddings:
    """
    class that will contains the embeddings for a given ontology
    """

    def __init__(self, BiomedicalOntology):

        alias_to_cuis: Dict[str, Set[str]] = defaultdict(set)
        self.cui_to_entity: Dict[str, BiomedicalEntity] = BiomedicalOntology.entities

        for entity in BiomedicalOntology.entities.values():
            if isinstance(entity.aliases, list):
                # If aliases is a list, use it directly : UMLS
                unique_aliases = entity.aliases
            elif isinstance(entity.aliases, str):
                # If aliases is a string, split using ";" or "|" : medic / entrez / mesh
                if entity.aliases.strip():  # Only split if the string is not empty
                    unique_aliases = re.split(r"[;|]", entity.aliases)
            else:
                # If aliases is None or any other type, use an empty list
                unique_aliases = []
            unique_aliases.append(entity.name)
            for alias in unique_aliases:
                alias_to_cuis[alias].add(entity.cui)

        self.alias_to_cuis: Dict[str, Set[str]] = {**alias_to_cuis}

        self.tfidf_vectorizer_path = None
        self.ann_index_path = None
        self.tfidf_vectors_path = None
        self.umls_concept_aliases_path = None

        # self.alias_to_cuis2: Dict[str, Set[str]] = {
        #     alias: list(cuis) for alias, cuis in alias_to_cuis.items()
        # }

        # with open(
        #     "/home2/cye73/data_test2/scispacy/umls_concept_aliases.json", "w"
        # ) as f:
        #     json.dump(self.alias_to_cuis2, f)

    # For Scispacy : create the embeddings and save them in different files in the provided folder path
    def create_tfidf_ann_index(
        self, out_path: str
    ) -> Tuple[List[str], TfidfVectorizer, FloatIndex]:
        """
        Build tfidf vectorizer and ann index.

        Parameters
        ----------
        out_path: str, required.
            The path where the various model pieces will be saved.
        kb : KnowledgeBase, optional.
            The kb items to generate the index and vectors for.

        """
        tfidf_vectorizer_path = f"{out_path}/tfidf_vectorizer.joblib"
        ann_index_path = f"{out_path}/nmslib_index.bin"
        tfidf_vectors_path = f"{out_path}/tfidf_vectors_sparse.npz"
        umls_concept_aliases_path = f"{out_path}/concept_aliases.json"

        self.tfidf_vectorizer_path = tfidf_vectorizer_path
        self.ann_index_path = ann_index_path
        self.tfidf_vectors_path = tfidf_vectors_path
        self.umls_concept_aliases_path = umls_concept_aliases_path

        # kb = kb or UmlsKnowledgeBase()

        # nmslib hyperparameters (very important)
        # guide: https://github.com/nmslib/nmslib/blob/master/manual/methods.md
        # Default values resulted in very low recall.

        # set to the maximum recommended value. Improves recall at the expense of longer indexing time.
        # We use the HNSW (Hierarchical Navigable Small World Graph) representation which is constructed
        # by consecutive insertion of elements in a random order by connecting them to M closest neighbours
        # from the previously inserted elements. These later become bridges between the network hubs that
        # improve overall graph connectivity. (bigger M -> higher recall, slower creation)
        # For more details see:  https://arxiv.org/pdf/1603.09320.pdf?
        m_parameter = 100
        # `C` for Construction. Set to the maximum recommended value
        # Improves recall at the expense of longer indexing time
        construction = 2000
        num_threads = 60  # set based on the machine
        index_params = {
            "M": m_parameter,
            "indexThreadQty": num_threads,
            "efConstruction": construction,
            "post": 0,
        }

        print(
            f"No tfidf vectorizer on {tfidf_vectorizer_path} or ann index on {ann_index_path}"
        )
        concept_aliases = list(self.alias_to_cuis.keys())

        # Filter out None values and empty strings
        concept_aliases = [
            alias.strip()
            for alias in concept_aliases
            if alias is not None and alias.strip()
        ]

        # NOTE: here we are creating the tf-idf vectorizer with float32 type, but we can serialize the
        # resulting vectors using float16, meaning they take up half the memory on disk. Unfortunately
        # we can't use the float16 format to actually run the vectorizer, because of this bug in sparse
        # matrix representations in scipy: https://github.com/scipy/scipy/issues/7408
        print(f"Fitting tfidf vectorizer on {len(concept_aliases)} aliases")
        tfidf_vectorizer = TfidfVectorizer(
            analyzer="char_wb", ngram_range=(3, 3), min_df=10, dtype=numpy.float32
        )
        start_time = datetime.datetime.now()
        concept_alias_tfidfs = tfidf_vectorizer.fit_transform(concept_aliases)
        print(f"Saving tfidf vectorizer to {tfidf_vectorizer_path}")
        joblib.dump(tfidf_vectorizer, tfidf_vectorizer_path)
        end_time = datetime.datetime.now()
        total_time = end_time - start_time
        print(
            f"Fitting and saving vectorizer took {total_time.total_seconds()} seconds"
        )

        print("Finding empty (all zeros) tfidf vectors")
        empty_tfidfs_boolean_flags = numpy.array(
            concept_alias_tfidfs.sum(axis=1) != 0
        ).reshape(-1)
        number_of_non_empty_tfidfs = sum(
            empty_tfidfs_boolean_flags == False
        )  # noqa: E712
        total_number_of_tfidfs = numpy.size(concept_alias_tfidfs, 0)

        print(
            f"Deleting {number_of_non_empty_tfidfs}/{total_number_of_tfidfs} aliases because their tfidf is empty"
        )
        # remove empty tfidf vectors, otherwise nmslib will crash
        concept_aliases = [
            alias
            for alias, flag in zip(concept_aliases, empty_tfidfs_boolean_flags)
            if flag
        ]
        concept_alias_tfidfs = concept_alias_tfidfs[empty_tfidfs_boolean_flags]
        assert len(concept_aliases) == numpy.size(concept_alias_tfidfs, 0)

        print(
            f"Saving list of concept ids and tfidfs vectors to {umls_concept_aliases_path} and {tfidf_vectors_path}"
        )
        json.dump(concept_aliases, open(umls_concept_aliases_path, "w"))
        scipy.sparse.save_npz(
            tfidf_vectors_path, concept_alias_tfidfs.astype(numpy.float16)
        )

        print(f"Fitting ann index on {len(concept_aliases)} aliases (takes 2 hours)")
        start_time = datetime.datetime.now()
        ann_index = nmslib.init(
            method="hnsw",
            space="cosinesimil_sparse",
            data_type=nmslib.DataType.SPARSE_VECTOR,
        )
        ann_index.addDataPointBatch(concept_alias_tfidfs)
        ann_index.createIndex(index_params, print_progress=True)
        ann_index.saveIndex(ann_index_path)
        end_time = datetime.datetime.now()
        elapsed_time = end_time - start_time
        print(f"Fitting ann index took {elapsed_time.total_seconds()} seconds")

        return concept_aliases, tfidf_vectorizer, ann_index

    def serialized_kb(self) -> LinkerPaths:
        tupled_kb = LinkerPaths(
            ann_index=self.ann_index_path,  # noqa
            tfidf_vectorizer=self.tfidf_vectorizer_path,  # noqa
            tfidf_vectors=self.tfidf_vectors_path,  # noqa
            concept_aliases_list=self.umls_concept_aliases_path,  # noqa
            alias_to_cuis=self.alias_to_cuis,
            cui_to_entity=self.cui_to_entity,
        )
        return tupled_kb


## Used to load the ann indexes from the binary file generated by create_tfidf_ann_index
def load_approximate_nearest_neighbours_index(
    serialized_kb: LinkerPaths,
    ef_search: int = 200,
) -> FloatIndex:
    """
    Load an approximate nearest neighbours index from disk.

    Parameters
    ----------
    linker_paths: LinkerPaths, required.
        Contains the paths to the data required for the entity linker.
    ef_search: int, optional (default = 200)
        Controls speed performance at query time. Max value is 2000,
        but reducing to around ~100 will increase query speed by an order
        of magnitude for a small performance hit.
    """
    concept_alias_tfidfs = scipy.sparse.load_npz(get_path(serialized_kb[2])).astype(
        numpy.float32
    )
    ann_index = nmslib.init(
        method="hnsw",
        space="cosinesimil_sparse",
        data_type=nmslib.DataType.SPARSE_VECTOR,
    )
    ann_index.addDataPointBatch(concept_alias_tfidfs)
    ann_index.loadIndex(get_path(serialized_kb[0]))
    query_time_params = {"efSearch": ef_search}
    ann_index.setQueryTimeParams(query_time_params)

    return ann_index



#cython: language_level=3
#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION

import cython
import numpy as np
cimport numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import minimum_spanning_tree
from tqdm import tqdm
from IPython import embed


INT = int
BOOL = bool
ctypedef np.int32_t INT_t
ctypedef np.float64_t FLOAT_t
ctypedef np.npy_bool BOOL_t


@cython.boundscheck(False)
@cython.wraparound(False)
def _build_adj_index(np.ndarray[INT_t, ndim=1] values,
                              INT_t max_value):
    # Required: values in ascending order
    cdef INT_t index_size = max_value + 1
    cdef np.ndarray[INT_t, ndim=2] adj_index = np.zeros([index_size, 2], dtype=INT)
    cdef INT_t i = 0, v
    cdef INT_t curr = values[0]

    for i, v in enumerate(values):
        if v != curr:
            curr = v
            adj_index[curr, 0] = i
            adj_index[curr, 1] = i + 1
        else:
            adj_index[curr, 1] += 1

    return adj_index

@cython.wraparound(False)
@cython.boundscheck(False)
def _EXP_has_entity_in_component(list stack,
                             FLOAT_t min_score,
                             np.ndarray[INT_t, ndim=1] to_vertices,
                             np.ndarray[FLOAT_t, ndim=1] weights,
                             np.ndarray[INT_t, ndim=2] adj_index,
                             INT_t num_entities,
                             bint dfs,
                             bint get_maximum,
                             bint debug):
    # Perform DFS to look for an entity
    cdef set visited = set()
    cdef bint found = False
    cdef INT_t node
    cdef INT_t immediate_child = -1
    cdef INT_t best_immediate_child = -1
    cdef FLOAT_t path_score = 1
    cdef FLOAT_t node_wt
    cdef FLOAT_t best_found_score = -1
    cdef list weight_stack = [1]
    cdef list prvs_score_stack = []
    cdef list nchild_stack = [1]
    cdef INT_t nchild_popped
    cdef INT_t last_idx
    cdef INT_t go_to_parent = 2 

    while len(stack) > 0:
        if debug:
            print("1")
        # Pop
        if dfs:
            if nchild_stack[len(nchild_stack) - 1] == 0: # go to parent's next sibling
                if debug:
                    print(f'stack={stack}')
                    print(f'prvs_score_stack={prvs_score_stack}')
                    print(f'nchild_stack={nchild_stack}')
                nchild_stack = nchild_stack[:len(nchild_stack) - 1]
                if debug:
                    print("1a")
                path_score = prvs_score_stack[len(prvs_score_stack) - go_to_parent]
                if debug:
                    print("1b")
                prvs_score_stack = prvs_score_stack[:len(prvs_score_stack) - go_to_parent]
                if debug:
                    print("1c")
            if debug:
                print("2")
            # DFS
            last_idx = len(stack) - 1
            node = stack[last_idx]
            # Track the immediate edge from the root of the current DFS path being evaluated
            if len(nchild_stack) == 2:
                immediate_child = node
            stack = stack[:last_idx]
            node_wt = weight_stack[last_idx]
            weight_stack = weight_stack[:last_idx]
            # Skip iteration if node has already been visited
            if node in visited:
                nchild_stack[len(nchild_stack) - 1] = nchild_stack[len(nchild_stack) - 1] - 1
                go_to_parent = 1
                continue

            if path_score * node_wt < max(min_score, best_found_score):
                nchild_stack[len(nchild_stack) - 1] = nchild_stack[len(nchild_stack) - 1] - 1
                go_to_parent = 1
                continue
            else:
                prvs_score_stack.append(path_score)
                path_score = path_score * node_wt
                go_to_parent = 2
            if debug:
                print("3")
        else:
            # BFS
            node = stack[0]
            stack = stack[1:]

        # Check if node is an entity
        if node < num_entities:
            best_found_score = path_score
            best_immediate_child = immediate_child
            if debug:
                print("4")
            if not get_maximum:
                return best_found_score, best_immediate_child

        visited.add(node)
        if debug:
            print("5")
        # Push all nodes reachable from the current node onto the stack
        start_idx, end_idx = adj_index[node, 0], adj_index[node, 1]
        next_nodes = to_vertices[start_idx:end_idx].tolist()
        if len(next_nodes) != 0:
            stack.extend(next_nodes)
            weight_stack.extend(weights[start_idx:end_idx].tolist())
            nchild_stack.append(len(next_nodes))
            if debug:
                print("6")
        else:
            nchild_stack[len(nchild_stack) - 1] = nchild_stack[len(nchild_stack) - 1] - 1
            if nchild_stack[len(nchild_stack) - 1] != 0:
                path_score = prvs_score_stack[len(prvs_score_stack) - 1]
                prvs_score_stack = prvs_score_stack[:len(prvs_score_stack) - 1]
            if debug:
                print("7")
    return best_found_score, best_immediate_child

@cython.wraparound(False)
@cython.boundscheck(False)
def _has_entity_in_component(list stack,
                             np.ndarray[INT_t, ndim=1] to_vertices,
                             np.ndarray[INT_t, ndim=2] adj_index,
                             INT_t num_entities,
                             bint dfs):
    # Perform DFS to look for an entity
    cdef set visited = set()
    cdef bint found = False
    cdef INT_t node
    
    while len(stack) > 0:
        # Pop
        if dfs:
            # DFS
            node = stack[len(stack) - 1]
            stack = stack[:len(stack) - 1]
        else:
            # BFS
            node = stack[0]
            stack = stack[1:]

        # Check if node is an entity
        if node < num_entities:
            found = True
            break

        # Skip iteration if node has already been visited
        if node in visited:
            continue
        visited.add(node)

        # Push all nodes reachable from the current node onto the stack
        start_idx, end_idx = adj_index[node, 0], adj_index[node, 1]
        stack.extend(to_vertices[start_idx:end_idx].tolist())

    return found

@cython.boundscheck(False)
@cython.wraparound(False)
def _EXP_special_partition(np.ndarray[INT_t, ndim=1] row, 
                      np.ndarray[INT_t, ndim=1] col,
                      np.ndarray[FLOAT_t, ndim=1] data,
                      np.ndarray[INT_t, ndim=1] ordered_indices,
                      np.ndarray[INT_t, ndim=1] siamese_indices,
                      dict edge_indices,
                      INT_t num_entities,
                      bint directed,
                      bint dfs,
                      bint silent):
    cdef INT_t num_edges = row.shape[0]
    cdef np.ndarray[BOOL_t, ndim=1] keep_mask = np.ones([num_edges,], dtype=BOOL)
    cdef np.ndarray[INT_t, ndim=1] tmp_graph
    cdef INT_t r, c
    # Flags to track if an entity is reachable from the row or the column, respectively, of the edge that is being dropped
    cdef bint r_entity_reachable, c_entity_reachable = True

    # Build the adjacency matrix for efficient DFS; row-wise for directed, col-wise for undirected
    # Shape [N, 2]; [x,0] to [x,1] (exclusive) is the range of indices for x
    # Example (row adjacency): row X has outgoing edges from X to all values in col[adj_index[X,0]:adj_index[X,1]]
    cdef np.ndarray[INT_t, ndim=2] adj_index
    cdef INT_t max_value = row[len(row) - 1] if directed else col[len(col) - 1] # Last value is max because of sorting
    cdef bint reachable_from_edge
    cdef FLOAT_t score_from_edge
    cdef FLOAT_t r_entity_score
    
    # Scale data to [0, 1]
    data = (data - np.min(data)) / (np.max(data) - np.min(data))
    
    adj_index = _build_adj_index(row if directed else col, max_value)

    if silent:
        _iter = ordered_indices
    else:
        _iter = tqdm(ordered_indices, desc='Partitioning joint graph')
    
    for i in _iter:
        # Undirected: Skip iteration if the edge has already been dropped
        if keep_mask[i] == False:
            continue

        r, c, edge_wt = row[i], col[i], data[i]

        # Remove the forward edge
        keep_mask[i] = False
        # Undirected: Remove the reverse edge
        if not directed:
            keep_mask[siamese_indices[i]] = False

        # Reduce the range by 1 in the adjacency index to reflect the dropped edge
        adj_index[r:, :] -= 1
        adj_index[r, 0] += 1
        if not directed:
            adj_index[c:, :] -= 1
            adj_index[c, 0] += 1

        # Create a temporary graph based on the dropped edge
        tmp_graph = col[keep_mask] if directed else row[keep_mask]
        # Check score of path from current edge to an entity
        reachable_from_edge = _has_entity_in_component(
            [c], tmp_graph, adj_index, num_entities, dfs)
        score_from_edge, _ = _EXP_has_entity_in_component(
                [c], -1, tmp_graph, data[keep_mask], adj_index, num_entities, dfs, True, False) # r==2344431 and c==2340167
        assert (not reachable_from_edge and score_from_edge == -1) or (reachable_from_edge and score_from_edge != -1)
        # Check if an entity can still be reached from r and that has a path score greater than the current path score
        r_entity_reachable = _has_entity_in_component([r], tmp_graph, adj_index, num_entities, dfs)
        if not reachable_from_edge and not r_entity_reachable:
            print("Unreachable with and without")
        r_entity_score, _ = _EXP_has_entity_in_component(
            [r], -1, tmp_graph, data[keep_mask], adj_index, num_entities, dfs, False, False)
        assert (not r_entity_reachable and r_entity_score == -1) or (r_entity_reachable and r_entity_score != -1)
        r_entity_score, best_path_node = _EXP_has_entity_in_component(
            [r], edge_wt * score_from_edge, tmp_graph, data[keep_mask], adj_index, num_entities, dfs, False, False)
        r_entity_reachable = r_entity_score != -1
            
        # Undirected: Check if an entity can still be reached from c
        # if not directed:
        #     c_entity_reachable = _EXP_has_entity_in_component(
        #         [c], tmp_graph, adj_index, num_entities, dfs)

        # Add (r,c) back if an entity cannot be reached from r or c (when undirected) without it
        if not (r_entity_reachable and c_entity_reachable):
            keep_mask[i] = True
            adj_index[r:, :] += 1
            adj_index[r, 0] -= 1
            # Remove the best edge in the temporary graph that was rejected
            if best_path_node != -1:
                adj_index[r:, :] -= 1
                adj_index[r, 0] += 1
                keep_mask[edge_indices[(r,best_path_node)]] = False
            if not directed:
                keep_mask[siamese_indices[i]] = True
                adj_index[c:, :] += 1
                adj_index[c, 0] -= 1

    return keep_mask

@cython.boundscheck(False)
@cython.wraparound(False)
def special_partition(np.ndarray[INT_t, ndim=1] row, 
                      np.ndarray[INT_t, ndim=1] col,
                      np.ndarray[INT_t, ndim=1] ordered_indices,
                      np.ndarray[INT_t, ndim=1] siamese_indices,
                      INT_t num_entities,
                      bint directed,
                      bint dfs,
                      bint silent):
    cdef INT_t num_edges = row.shape[0]
    cdef np.ndarray[BOOL_t, ndim=1] keep_mask = np.ones([num_edges,], dtype=BOOL)
    cdef np.ndarray[INT_t, ndim=1] tmp_graph
    cdef INT_t r, c
    # Flags to track if an entity is reachable from the row or the column, respectively, of the edge that is being dropped
    cdef bint r_entity_reachable, c_entity_reachable = True

    # Build the adjacency matrix for efficient DFS; row-wise for directed, col-wise for undirected
    # Shape [N, 2]; [x,0] to [x,1] (exclusive) is the range of indices for x
    # Example (row adjacency): row X has outgoing edges from X to all values in col[adj_index[X,0]:adj_index[X,1]]
    cdef np.ndarray[INT_t, ndim=2] adj_index
    cdef INT_t max_value = row[len(row) - 1] if directed else col[len(col) - 1] # Last value is max because of sorting
    adj_index = _build_adj_index(row if directed else col, max_value)

    if silent:
        _iter = ordered_indices
    else:
        _iter = tqdm(ordered_indices, desc='Partitioning joint graph')
    
    for i in _iter:
        # Undirected: Skip iteration if the edge has already been dropped
        if keep_mask[i] == False:
            continue

        r, c = row[i], col[i]

        # Remove the forward edge
        keep_mask[i] = False
        # Undirected: Remove the reverse edge
        if not directed:
            keep_mask[siamese_indices[i]] = False

        # Reduce the range by 1 in the adjacency index to reflect the dropped edge
        adj_index[r:, :] -= 1
        adj_index[r, 0] += 1
        if not directed:
            adj_index[c:, :] -= 1
            adj_index[c, 0] += 1

        # Create a temporary graph based on the dropped edge
        tmp_graph = col[keep_mask] if directed else row[keep_mask]

        # Check if an entity can still be reached from r
        r_entity_reachable = _has_entity_in_component(
            [r], tmp_graph, adj_index, num_entities, dfs)
        # Undirected: Check if an entity can still be reached from c
        if not directed:
            c_entity_reachable = _has_entity_in_component(
                [c], tmp_graph, adj_index, num_entities, dfs)

        # Add (r,c) back if an entity cannot be reached from r or c (when undirected) without it
        if not (r_entity_reachable and c_entity_reachable):
            keep_mask[i] = True
            adj_index[r:, :] += 1
            adj_index[r, 0] -= 1
            if not directed:
                keep_mask[siamese_indices[i]] = True
                adj_index[c:, :] += 1
                adj_index[c, 0] -= 1

    return keep_mask

def cluster_linking_partition(rows, cols, data, n_entities, directed=True, dfs=True, silent=False, exclude=set(), threshold=None, experimental=False):
    assert rows.shape[0] == cols.shape[0] == data.shape[0]
    
    cdef np.ndarray[BOOL_t, ndim=1] keep_edge_mask

    # Filter duplicates only on row,col tuples (to accomodate approximation errors in data)
    # Additionally, filter out any nodes passed in 'exclude'
    seen = set()
    duplicated, excluded, thresholded = 0, 0, 0
    _f_row, _f_col, _f_data = [], [], []
    for k in range(len(rows)):
        if (rows[k], cols[k]) in seen:
            duplicated += 1
            continue
        seen.add((rows[k], cols[k]))
        if rows[k] in exclude or cols[k] in exclude:
            excluded += 1
            continue
        if threshold is not None and data[k] < threshold:
            thresholded += 1
            continue
        _f_row.append(rows[k])
        _f_col.append(cols[k])
        _f_data.append(data[k])
    rows, cols, data = list(map(np.array, (_f_row, _f_col, _f_data)))

    if duplicated + excluded + thresholded > 0:
        print(f"""
Dropped edges during pre-processing:
    Duplicates: {duplicated}
    Excluded: {excluded}
    Thresholded: {thresholded}""")

    if not directed:
        # Filter down using Scipy's MST routine for faster processing
        shape = int(max(np.max(rows), np.max(cols))) + 1
        shape = (shape, shape)
        csr = csr_matrix((-data, (rows, cols)), shape=shape)
        mst = minimum_spanning_tree(csr).tocoo()
        rows, cols, data = mst.row, mst.col, -mst.data
    
        # Add the reverse edges
        rows, cols = np.concatenate((rows, cols)), np.concatenate((cols, rows))
        data = np.concatenate((data, data))

        # Filter duplicates only on row,col tuples (to accomodate approximation errors in data)
        seen = set()
        _f_row, _f_col, _f_data = [], [], []
        for k in range(len(rows)):
            if (rows[k], cols[k]) in seen:
                continue
            seen.add((rows[k], cols[k]))
            _f_row.append(rows[k])
            _f_col.append(cols[k])
            _f_data.append(data[k])
        rows, cols, data = list(map(np.array, (_f_row, _f_col, _f_data)))

    # Sort data for efficient DFS
    sort_order = lambda x: (x[0], x[1]*(-1 if dfs else 1)) if directed else (x[1], x[0]*(-1 if dfs else 1)) # For faster iterations: descending order for DFS, ascending order for BFS
    tuples = zip(rows, cols, data)
    tuples = sorted(tuples, key=sort_order)
    rows, cols, data = zip(*tuples)
    rows = np.asarray(rows, dtype=INT)
    cols = np.asarray(cols, dtype=INT)
    data = np.asarray(data)

    # If undirected, create siamese indices for reverse lookup (i.e. c,r edge to index)
    cdef dict edge_idxs = {e: i for i, e in enumerate(zip(rows, cols))}
    siamese_idxs = np.array([edge_idxs[(r_c[1], r_c[0])] 
                            if (r_c[1], r_c[0]) in edge_idxs else -1
                            for r_c in edge_idxs])

    # Order the edges in ascending order of similarity scores
    ordered_edge_idxs = np.argsort(data)

    # Determine which edges to keep in the partitioned graph
    if experimental:
        keep_edge_mask = _EXP_special_partition(
            rows, cols, data, ordered_edge_idxs, siamese_idxs, edge_idxs, n_entities, directed, dfs, silent)
    else:
        keep_edge_mask = special_partition(
            rows, cols, ordered_edge_idxs, siamese_idxs, n_entities, directed, dfs, silent)

    # Return the edges of the partitioned graph
    return rows[keep_edge_mask], cols[keep_edge_mask], data[keep_edge_mask]


from transformers.models.bart.modeling_bart import *
import torch

_CONFIG_FOR_DOC = "BartConfig"

@add_start_docstrings(
    "The BART Model with a language modeling head. Can be used for summarization.", BART_START_DOCSTRING
)
class BartForConditionalGeneration(BartPretrainedModel):
    base_model_prefix = "model"
    _keys_to_ignore_on_load_missing = [r"final_logits_bias", r"lm_head\.weight"]

    def __init__(self, config: BartConfig):
        super().__init__(config)
        self.model = BartModel(config)
        self.register_buffer("final_logits_bias", torch.zeros((1, self.model.shared.num_embeddings)))
        self.lm_head = nn.Linear(config.d_model, self.model.shared.num_embeddings, bias=False)

        self.init_weights()

    def get_encoder(self):
        return self.model.get_encoder()

    def get_decoder(self):
        return self.model.get_decoder()

    def resize_token_embeddings(self, new_num_tokens: int) -> nn.Embedding:
        new_embeddings = super().resize_token_embeddings(new_num_tokens)
        self._resize_final_logits_bias(new_num_tokens)
        return new_embeddings

    def _resize_final_logits_bias(self, new_num_tokens: int) -> None:
        old_num_tokens = self.final_logits_bias.shape[-1]
        if new_num_tokens <= old_num_tokens:
            new_bias = self.final_logits_bias[:, :new_num_tokens]
        else:
            extra_bias = torch.zeros((1, new_num_tokens - old_num_tokens), device=self.final_logits_bias.device)
            new_bias = torch.cat([self.final_logits_bias, extra_bias], dim=1)
        self.register_buffer("final_logits_bias", new_bias)

    def get_output_embeddings(self):
        return self.lm_head

    def set_output_embeddings(self, new_embeddings):
        self.lm_head = new_embeddings

    @add_start_docstrings_to_model_forward(BART_INPUTS_DOCSTRING)
    @replace_return_docstrings(output_type=Seq2SeqLMOutput, config_class=_CONFIG_FOR_DOC)
    @add_end_docstrings(BART_GENERATION_EXAMPLE)
    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        decoder_input_ids=None,
        decoder_attention_mask=None,
        head_mask=None,
        decoder_head_mask=None,
        cross_attn_head_mask=None,
        encoder_outputs=None,
        past_key_values=None,
        inputs_embeds=None,
        decoder_inputs_embeds=None,
        labels=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        r"""
        labels (:obj:`torch.LongTensor` of shape :obj:`(batch_size, sequence_length)`, `optional`):
            Labels for computing the masked language modeling loss. Indices should either be in ``[0, ...,
            config.vocab_size]`` or -100 (see ``input_ids`` docstring). Tokens with indices set to ``-100`` are ignored
            (masked), the loss is only computed for the tokens with labels in ``[0, ..., config.vocab_size]``.
        Returns:
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if labels is not None and decoder_inputs_embeds is None:
            if decoder_input_ids is None:
                decoder_input_ids = shift_tokens_right(
                    labels, self.config.pad_token_id, self.config.decoder_start_token_id
                )

        outputs = self.model(
            input_ids,
            attention_mask=attention_mask,
            decoder_input_ids=decoder_input_ids,
            encoder_outputs=encoder_outputs,
            decoder_attention_mask=decoder_attention_mask,
            head_mask=head_mask,
            decoder_head_mask=decoder_head_mask,
            cross_attn_head_mask=cross_attn_head_mask,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            decoder_inputs_embeds=decoder_inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        lm_logits = self.lm_head(outputs[0]) + self.final_logits_bias

        masked_lm_loss = None
        if labels is not None:
            loss_fct = CrossEntropyLoss()
            masked_lm_loss = loss_fct(lm_logits.view(-1, self.config.vocab_size), labels.view(-1))

        if not return_dict:
            output = (lm_logits,) + outputs[1:]
            return ((masked_lm_loss,) + output) if masked_lm_loss is not None else output

        return Seq2SeqLMOutput(
            loss=masked_lm_loss,
            logits=lm_logits,
            past_key_values=outputs.past_key_values,
            decoder_hidden_states=outputs.decoder_hidden_states,
            decoder_attentions=outputs.decoder_attentions,
            cross_attentions=outputs.cross_attentions,
            encoder_last_hidden_state=outputs.encoder_last_hidden_state,
            encoder_hidden_states=outputs.encoder_hidden_states,
            encoder_attentions=outputs.encoder_attentions,
        )

    def prepare_inputs_for_generation(
        self,
        decoder_input_ids,
        past=None,
        attention_mask=None,
        head_mask=None,
        decoder_head_mask=None,
        cross_attn_head_mask=None,
        use_cache=None,
        encoder_outputs=None,
        **kwargs
    ):
        # cut decoder_input_ids if past is used
        if past is not None:
            decoder_input_ids = decoder_input_ids[:, -1:]

        return {
            "input_ids": None,  # encoder_outputs is defined. input_ids not needed
            "encoder_outputs": encoder_outputs,
            "past_key_values": past,
            "decoder_input_ids": decoder_input_ids,
            "attention_mask": attention_mask,
            "head_mask": head_mask,
            "decoder_head_mask": decoder_head_mask,
            "cross_attn_head_mask": cross_attn_head_mask,
            "use_cache": use_cache,  # change this to avoid caching (presumably for debugging)
        }

    def prepare_decoder_input_ids_from_labels(self, labels: torch.Tensor):
        return shift_tokens_right(labels, self.config.pad_token_id, self.config.decoder_start_token_id)

    @staticmethod
    def _reorder_cache(past, beam_idx):
        reordered_past = ()
        for layer_past in past:
            # cached cross_attention states don't have to be reordered -> they are always the same
            reordered_past += (
                tuple(past_state.index_select(0, beam_idx) for past_state in layer_past[:2]) + layer_past[2:],
            )
        return reordered_past




import os
import pickle
import logging
import time
import glob
import math


import multiprocessing
from typing import List, Tuple, Dict, Iterator, Set
from functools import partial
from multiprocessing.dummy import Pool

import numpy as np
import faiss

import torch
import torch.nn as nn
from torch import Tensor as T
from transformers import (
    AutoConfig,
    AutoTokenizer,
    AutoModel,
    PreTrainedTokenizer,
)

from bioel.models.krissbert.data.utils import BigBioDataset, generate_vectors

# Setup logger
logger = logging.getLogger()
logger.setLevel(logging.INFO)
log_formatter = logging.Formatter(
    "[%(thread)s] %(asctime)s [%(levelname)s] %(name)s: %(message)s"
)
console = logging.StreamHandler()
console.setFormatter(log_formatter)
logger.addHandler(console)

class DenseIndexer(object):
    def __init__(self, buffer_size: int = 50000):
        self.buffer_size = buffer_size
        self.index_id_to_db_id = []
        self.index = None

    def init_index(self, vector_sz: int):
        raise NotImplementedError

    def index_data(self, data: List[Tuple[object, np.array]]):
        raise NotImplementedError

    def get_index_name(self):
        raise NotImplementedError

    def search_knn(
        self, query_vectors: np.array, top_docs: int
    ) -> List[Tuple[List[object], List[float]]]:
        raise NotImplementedError

    def serialize(self, file: str):
        logger.info("Serializing index to %s", file)

        if os.path.isdir(file):
            index_file = os.path.join(file, "index.dpr")
            meta_file = os.path.join(file, "index_meta.dpr")
        else:
            index_file = file + ".index.dpr"
            meta_file = file + ".index_meta.dpr"

        faiss.write_index(self.index, index_file)
        with open(meta_file, mode="wb") as f:
            pickle.dump(self.index_id_to_db_id, f)

    def get_files(self, path: str):
        if os.path.isdir(path):
            index_file = os.path.join(path, "index.dpr")
            meta_file = os.path.join(path, "index_meta.dpr")
        else:
            index_file = path + ".index.dpr"
            meta_file = path + ".index_meta.dpr"
        return index_file, meta_file

    def index_exists(self, path: str):
        index_file, meta_file = self.get_files(path)
        return os.path.isfile(index_file) and os.path.isfile(meta_file)

    def deserialize(self, path: str):
        logger.info("Loading index from %s", path)
        index_file, meta_file = self.get_files(path)

        self.index = faiss.read_index(index_file)
        logger.info(
            "Loaded index of type %s and size %d", type(self.index), self.index.ntotal
        )

        with open(meta_file, "rb") as reader:
            self.index_id_to_db_id = pickle.load(reader)
        assert (
            len(self.index_id_to_db_id) == self.index.ntotal
        ), "Deserialized index_id_to_db_id should match faiss index size"

    def _update_id_mapping(self, db_ids: List) -> int:
        self.index_id_to_db_id.extend(db_ids)
        return len(self.index_id_to_db_id)


class DenseFlatIndexer(DenseIndexer):
    def __init__(self, buffer_size: int = 50000):
        super(DenseFlatIndexer, self).__init__(buffer_size=buffer_size)

    def init_index(self, vector_sz: int):
        self.index = faiss.IndexFlatIP(vector_sz)

    def index_data(self, data: List[Tuple[object, np.array]]):
        n = len(data)
        # indexing in batches is beneficial for many faiss index types
        for i in range(0, n, self.buffer_size):
            db_ids = [t[0] for t in data[i : i + self.buffer_size]]
            vectors = [
                np.reshape(t[1], (1, -1)) for t in data[i : i + self.buffer_size]
            ]
            vectors = np.concatenate(vectors, axis=0)
            total_data = self._update_id_mapping(db_ids)
            self.index.add(vectors)
            logger.info("data indexed %d", total_data)

        indexed_cnt = len(self.index_id_to_db_id)
        logger.info("Total data indexed %d", indexed_cnt)

    def search_knn(
        self,
        query_vectors: np.array,
        top_docs: int,
        batch_size: int = 4096,
    ) -> List[Tuple[List[object], List[float]]]:
        num_queries = query_vectors.shape[0]
        scores, indexes = [], []
        for start in range(0, num_queries, batch_size):
            logger.info(f"Searched {start} queries.")
            batch_vectors = query_vectors[start : start + batch_size]
            batch_scores, batch_indexes = self.index.search(batch_vectors, top_docs)
            scores.extend(batch_scores)
            indexes.extend(batch_indexes)
        # convert to external ids
        db_ids = [
            [self.index_id_to_db_id[i] for i in query_top_idxs]
            for query_top_idxs in indexes
        ]
        result = [(db_ids[i], scores[i]) for i in range(len(db_ids))]
        return result

    def get_index_name(self):
        return "flat_index"


def iterate_encoded_files(
    vector_files: list,
    candidate_ids: Set = None,
    umls_data: Dict = None,
) -> Iterator:
    logger.info("Loading encoded prototype embeddings...")
    proto_data = {}
    for file in vector_files:
        logger.info("Reading file %s", file)
        with open(file, "rb") as reader:
            for meta, vec in pickle.load(reader):
                cuis = meta["cuis"]
                if candidate_ids and all(c not in candidate_ids for c in cuis):
                    continue
                for cui in cuis:
                    proto_data.setdefault(cui, []).append((meta, vec))
    # Concatenate prototype embs with additional knowledge embs from UMLS.
    if umls_data is not None:
        for cui, (meta, vec) in umls_data.items():
            if cui in proto_data:
                for _, _vec in proto_data.pop(cui):
                    extended_vec = np.concatenate((vec, _vec), axis=0)
                    yield (meta, extended_vec)
            else:
                extended_vec = np.concatenate((vec, np.zeros_like(vec)), axis=0)
                yield (meta, extended_vec)
    for cui in list(proto_data.keys()):
        for meta, vec in proto_data.pop(cui):
            extended_vec = np.concatenate((np.zeros_like(vec), vec), axis=0)
            yield (meta, extended_vec)
    assert len(proto_data) == 0

class Krissbert(nn.Module):
    def __init__(self, model_name_or_path="microsoft/BiomedNLP-KRISSBERT-PubMed-UMLS-EL"):
        super().__init__()
        self.config = AutoConfig.from_pretrained(model_name_or_path)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name_or_path, use_fast=True)
        self.encoder = AutoModel.from_pretrained(model_name_or_path, config=self.config)
        self.index = DenseFlatIndexer()
        self.index.init_index(self.config.hidden_size * 2)

    def forward(self, input_ids, attention_mask):
        return self.encoder(input_ids=input_ids, attention_mask=attention_mask)

    def generate_prototypes(self, dataset, output_dir, batch_size=256, max_length=64):
        self.encoder.eval()
        data = generate_vectors(self.encoder, self.tokenizer, dataset, batch_size, max_length, is_prototype=True)
        output_dir = output_dir.rstrip("/")
        if isinstance(dataset, BigBioDataset):
            output_prototypes = f'{output_dir}/{dataset.dataset_name}_embeddings.pickle'
            output_name_cuis = f'{output_dir}/{dataset.dataset_name}_name_cuis.txt'
        else:
            output_prototypes = f'{output_dir}/embeddings.pickle'
            output_name_cuis = f'{output_dir}/name_cuis.txt'

        os.makedirs(os.path.dirname(output_prototypes), exist_ok=True)
        logger.info("Writing results to %s", output_prototypes)
        with open(output_prototypes, mode="wb") as f:
            pickle.dump(data, f)
        with open(output_name_cuis, "w") as f:
            for name, cuis in dataset.name_to_cuis.items():
                f.write("|".join(cuis) + "||" + name + "\n")
        logger.info("Total data processed %d. Written to %s", len(data), output_prototypes)

    def generate_mention_vectors(self, ds, batch_size=256, max_length=64):
        self.encoder.eval()
        return generate_vectors(
            encoder=self.encoder,
            tokenizer=self.tokenizer,
            dataset=ds,
            batch_size=batch_size,
            max_length=max_length,
        )
    
    def index_encoded_data(
        self,
        vector_files: List[str],
        buffer_size: int,
        candidate_ids: Set = None,
        umls_data: Dict = None,
    ):
        """
        Indexes encoded data takes form a list of files
        :param vector_files: a list of files
        :param buffer_size: size of a buffer to send for the indexing at once
        :return:
        """
        buffer = []
        for i, item in enumerate(
            iterate_encoded_files(vector_files, candidate_ids, umls_data)
        ):
            buffer.append(item)
            if 0 < buffer_size == len(buffer):
                self.index.index_data(buffer)
                buffer = []
        self.index.index_data(buffer)
        logger.info("Data indexing completed.")

    def get_top_hits(
        self, mention_vectors: np.array, top_k: int = 100
    ) -> List[Tuple[List[object], List[float]]]:
        """
        Does the retrieval of the best matching given the mention vectors batch
        """
        time0 = time.time()
        search = partial(
            self.index.search_knn,
            top_docs=top_k,
        )
        results = []
        num_processes = multiprocessing.cpu_count()
        shard_size = math.ceil(mention_vectors.shape[0] / num_processes)
        shards = []
        for i in range(0, mention_vectors.shape[0], shard_size):
            shards.append(mention_vectors[i : i + shard_size])
        with Pool(processes=num_processes) as pool:
            it = pool.map(search, shards)
            for ret in it:
                results += ret
            # results = self.index.search_knn(mention_vectors, top_k)
        logger.info("index search time: %f sec.", time.time() - time0)
        self.index = None
        return results


import argparse
import logging
import os
import torch
import pickle
import ujson
from tqdm.auto import tqdm

from bioel.models.sapbert.data.utils import DictionaryDataset


LOGGER = logging.getLogger()


def parse_args_bigbio():
    """
    Parse input arguments
    """
    parser = argparse.ArgumentParser(description="sapbert evaluation")

    # Required
    parser.add_argument("--model_dir", required=True, help="Directory for model")
    parser.add_argument(
        "--dictionary_path", type=str, required=True, help="dictionary path"
    )
    parser.add_argument(
        "--dataset_name",
        type=str,
        required=True,
        choices=[
            "medmentions_full",
            "medmentions_st21pv",
            "bc5cdr",
            "nlmchem",
            "nlm_gene",
            "gnormplus",
            "ncbi_disease",
        ],
        help="data set to evaluate",
    )
    parser.add_argument("--dict_cache_path", type=str)

    # Run settings
    parser.add_argument("--use_cuda", action="store_true")
    parser.add_argument(
        "--output_dir", type=str, default="./output/", help="Directory for output"
    )
    parser.add_argument("--batch_size", type=int, default=32)
    parser.add_argument("--dict_batch_size", type=int, default=8192)
    parser.add_argument("--topk", type=int, default=10)
    parser.add_argument(
        "--split",
        default="test",
        type=str,
        help="Which datasplit to evaluate model performance on",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Run model with small data subset as a test of functionality",
    )

    parser.add_argument(
        "--abbreviations_path",
        type=str,
        help="Path to .json file of abbreviations in each article",
    )

    # Tokenizer settings
    parser.add_argument("--max_length", default=25, type=int)

    # Classification settings
    parser.add_argument(
        "--agg_mode", type=str, default="cls", help="{cls|mean_pool|nospec}"
    )

    args = parser.parse_args()
    return args


def make_unique_model_savepath(dir, model, dataset, file_ext):
    # Ensure the directory exists
    if not os.path.exists(dir):
        os.makedirs(dir)

    # Create the base filename
    base_filename = f"{model}_{dataset}.{file_ext}"
    unique_filename = base_filename
    counter = 1

    # Append a number to the filename until a unique one is found
    while os.path.exists(os.path.join(dir, unique_filename)):
        unique_filename = f"{model}_{dataset}_{counter}.{file_ext}"
        counter += 1

    return os.path.join(dir, unique_filename)


def init_logging():
    LOGGER.setLevel(logging.INFO)
    fmt = logging.Formatter("%(asctime)s: [ %(message)s ]", "%m/%d/%Y %I:%M:%S %p")
    console = logging.StreamHandler()
    console.setFormatter(fmt)
    LOGGER.addHandler(console)


def load_dictionary(dictionary_path):
    dictionary = DictionaryDataset(dictionary_path=dictionary_path)
    return dictionary.data


def sapbert_predict(
    model_wrapper,
    eval_dictionary,
    dataloader,
    batch_size: int = 64,
    dict_batch_size: int = 8192,
    topk: int = 25,
    agg_mode: str = "cls",
    dict_cache_filepath=None,
    db_name: str = "UMLS",
    debug=False,
):
    """
    Make predictions for SapBERT given a particular dict of medical concepts
    """
    # # Quick test of dataloader
    # for i, dat in enumerate(dataloader):
    #     print(ujson.dumps(dat))

    print("[start embedding dictionary...]")
    # embed dictionary
    dict_names = [row[0] for row in eval_dictionary]

    if debug:
        dict_names = dict_names[:4096]

    if dict_cache_filepath is not None:
        if os.path.isfile(dict_cache_filepath):
            print(f"Found cached dict at {dict_cache_filepath}.  Loading.")
            dict_dense_embeds = torch.load(
                dict_cache_filepath,
            )
        else:
            dict_dense_embeds = model_wrapper.embed_dense(
                names=dict_names,
                show_progress=True,
                batch_size=dict_batch_size,
                agg_mode=agg_mode,
            )
            print(f"Saving dict to {dict_cache_filepath}")
            torch.save(
                dict_dense_embeds,
                dict_cache_filepath,
                pickle_protocol=pickle.HIGHEST_PROTOCOL,
            )
    else:
        dict_dense_embeds = model_wrapper.embed_dense(
            names=dict_names,
            show_progress=True,
            batch_size=dict_batch_size,
            agg_mode=agg_mode,
        )

    # print ("dict_dense_embeds.shape:", dict_dense_embeds.shape)

    print("[computing rankings...]")

    results = []
    for batch in tqdm(dataloader):
        # for i in tqdm(np.arange(0,len(eval_queries),batch_size), total=len(eval_queries)//batch_size+1):
        # mentions = list(eval_queries[i:i+batch_size][:,0])
        mentions = batch[0]
        mention_dense_embeds = model_wrapper.embed_dense(
            names=mentions, agg_mode=agg_mode
        )
        metadata = batch[-1]

        # get score matrix
        dense_score_matrix = model_wrapper.get_score_matrix(
            query_embeds=mention_dense_embeds, dict_embeds=dict_dense_embeds
        )
        score_matrix = dense_score_matrix
        candidate_idxs_batch = model_wrapper.retrieve_candidate_cuda(
            score_matrix=score_matrix,
            topk=topk,
            batch_size=batch_size,
            show_progress=False,
        )

        for i, idx_list in enumerate(candidate_idxs_batch):
            np_candidates = [eval_dictionary[ind] for ind in idx_list.tolist()]
            # print(np_candidates)
            candidates_extra = [{"text": x[0], "db_id": x[1]} for x in np_candidates]
            candidates = [x[1].split("|") for x in np_candidates]
            metadata[i]["candidates"] = candidates
            metadata[i]["candidates_metadata"] = candidates_extra
            if len(results) == 0 and debug:
                LOGGER.debug(metadata)
        results.extend(metadata)
        # print(f"results : {results[:10]}")

        if debug:
            break

    return results



from bioel.models.krissbert.data.utils import BigBioDataset
import json
import torch
import os


def train_model(params, model):

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print(f"Using device: {device}")
    model.to(device)

    ds = BigBioDataset(
        params["dataset_name"],
        splits=["train"],
        abbreviations_path=params["path_to_abbrev"],
    )

    model.generate_prototypes(
        ds, params["output_dir"], params["batch_size"], params["max_length"]
    )

    print("Done generating prototypes!")



import os
import pickle
import logging
import torch
import numpy as np
import time
from tqdm import tqdm
from torch import nn
from sklearn.metrics.pairwise import cosine_similarity
from .metric_learning import *

from transformers import (
    AutoTokenizer,
    AutoModel,
)

LOGGER = logging.getLogger()


class Model_Wrapper(object):
    """
    Wrapper class for BERT encoder
    """

    def __init__(self):
        self.tokenizer = None
        self.encoder = None

    def get_dense_encoder(self):
        assert self.encoder is not None

        return self.encoder

    def get_dense_tokenizer(self):
        assert self.tokenizer is not None

        return self.tokenizer

    def save_model(self, path, context=False):
        # save bert model, bert config
        LOGGER.info(f"Saving the model to {path}")
        for param in self.encoder.parameters():
            if not param.is_contiguous():
                param.data = param.data.contiguous()
        self.encoder.save_pretrained(path)

        # save bert vocab
        self.tokenizer.save_pretrained(path)

    def load_model(self, path, max_length=25, use_cuda=True, lowercase=True):
        self.load_bert(path, max_length, use_cuda)

        return self

    def load_bert(self, path, max_length, use_cuda, lowercase=True):
        self.tokenizer = AutoTokenizer.from_pretrained(
            path, use_fast=True, do_lower_case=lowercase
        )
        self.encoder = AutoModel.from_pretrained(path)
        if use_cuda:
            self.encoder = self.encoder.cuda()

        return self.encoder, self.tokenizer

    def load_pretrained_model(self, path, use_cuda=True):
        # Load the saved model
        self.encoder = AutoModel.from_pretrained(path)

        if use_cuda:
            self.encoder = self.encoder.cuda()

    def get_score_matrix(
        self, query_embeds, dict_embeds, cosine=False, normalise=False
    ):
        """
        Return score matrix

        Parameters
        ----------
        query_embeds : np.array
            2d numpy array of query embeddings
        dict_embeds : np.array
            2d numpy array of query embeddings

        Returns
        -------
        score_matrix : np.array
            2d numpy array of scores
        """
        if cosine:
            score_matrix = cosine_similarity(query_embeds, dict_embeds)
        else:
            score_matrix = np.matmul(query_embeds, dict_embeds.T)

        if normalise:
            score_matrix = (score_matrix - score_matrix.min()) / (
                score_matrix.max() - score_matrix.min()
            )

        return score_matrix

    def retrieve_candidate(self, score_matrix, topk):
        """
        Return sorted topk idxes (descending order)

        Parameters
        ----------
        score_matrix : np.array
            2d numpy array of scores
        topk : int
            The number of candidates

        Returns
        -------
        topk_idxs : np.array
            2d numpy array of scores [# of query , # of dict]
        """

        def indexing_2d(arr, cols):
            rows = np.repeat(
                np.arange(0, cols.shape[0])[:, np.newaxis], cols.shape[1], axis=1
            )
            return arr[rows, cols]

        # get topk indexes without sorting
        topk_idxs = np.argpartition(score_matrix, -topk)[:, -topk:]

        # get topk indexes with sorting
        topk_score_matrix = indexing_2d(score_matrix, topk_idxs)
        topk_argidxs = np.argsort(-topk_score_matrix)
        topk_idxs = indexing_2d(topk_idxs, topk_argidxs)

        return topk_idxs

    def retrieve_candidate_cuda(
        self, score_matrix, topk, batch_size=128, show_progress=False
    ):
        """
        Return sorted topk idxes (descending order)

        Parameters
        ----------
        score_matrix : np.array
            2d numpy array of scores
        topk : int
            The number of candidates

        Returns
        -------
        topk_idxs : np.array
            2d numpy array of scores [# of query , # of dict]
        """

        res = None
        for i in tqdm(
            np.arange(0, score_matrix.shape[0], batch_size), disable=not show_progress
        ):
            score_matrix_tmp = torch.tensor(score_matrix[i : i + batch_size]).cuda()
            matrix_sorted = torch.argsort(score_matrix_tmp, dim=1, descending=True)[
                :, :topk
            ].cpu()
            if res is None:
                res = matrix_sorted
            else:
                res = torch.cat([res, matrix_sorted], axis=0)

        return res.numpy()

    def embed_dense(self, names, show_progress=False, batch_size=2048, agg_mode="cls"):
        """
        Embedding data into dense representations

        Parameters
        ----------
        names : np.array
            An array of names

        Returns
        -------
        dense_embeds : list
            A list of dense embeddings
        """
        self.encoder.eval()  # prevent dropout

        batch_size = batch_size
        dense_embeds = []

        # print ("converting names to list...")
        # names = names.tolist()

        with torch.no_grad():
            if show_progress:
                iterations = tqdm(range(0, len(names), batch_size))
            else:
                iterations = range(0, len(names), batch_size)

            for start in iterations:
                end = min(start + batch_size, len(names))
                batch = names[start:end]
                batch_tokenized_names = self.tokenizer.batch_encode_plus(
                    batch,
                    add_special_tokens=True,
                    truncation=True,
                    max_length=25,
                    padding="max_length",
                    return_tensors="pt",
                )
                batch_tokenized_names_cuda = {}
                for k, v in batch_tokenized_names.items():
                    batch_tokenized_names_cuda[k] = v.cuda()

                last_hidden_state = self.encoder(**batch_tokenized_names_cuda)[0]
                if agg_mode == "cls":
                    batch_dense_embeds = last_hidden_state[:, 0, :]  # [CLS]
                elif agg_mode == "mean_all_tok":
                    batch_dense_embeds = last_hidden_state.mean(1)  # pooling
                elif agg_mode == "mean":
                    batch_dense_embeds = (
                        last_hidden_state
                        * batch_tokenized_names_cuda["attention_mask"].unsqueeze(-1)
                    ).sum(1) / batch_tokenized_names_cuda["attention_mask"].sum(
                        -1
                    ).unsqueeze(
                        -1
                    )
                else:
                    print("no such agg_mode:", agg_mode)

                batch_dense_embeds = batch_dense_embeds.cpu().detach().numpy()
                dense_embeds.append(batch_dense_embeds)
        dense_embeds = np.concatenate(dense_embeds, axis=0)

        return dense_embeds



import os
import ujson
import warnings
from collections import defaultdict
import pandas as pd

# from bigbio.dataloader import BigBioConfigHelpers
import json
import joblib
import numpy as np
import datetime

from tqdm.auto import tqdm

tqdm.pandas()

from datasets import load_dataset
from bioel.utils.bigbio_utils import (
    dataset_to_documents,
    dataset_to_df,
    resolve_abbreviation,
    load_bigbio_dataset,
)
from bioel.utils.dataset_consts import (
    CUIS_TO_REMAP,
    CUIS_TO_EXCLUDE,
    VALIDATION_DOCUMENT_IDS,
)
from bioel.ontology import BiomedicalOntology
from sklearn.feature_extraction.text import TfidfVectorizer

from bioel.ontology import BiomedicalOntology
from typing import List


def cuis_to_aliases(ontology: BiomedicalOntology, save_dir: str, dataset_name: str):
    """
    Creates the .txt file that maps the Cuis to the aliases and canonical name for each entities
    from the provided ontology.

    Parameters
    ----------
    ontology_dir: str, required.
        The path where the ontology is stored.
    save_dir : str, required.
        The path where the mapping file "{dataset_name}_aliases.txt" will be stored
    dataset_name : str, required.
        The name of the dataset for which the mapping will be created.
    """
    # Create the directory if it doesn't exist
    os.makedirs(save_dir, exist_ok=True)
    import re

    file_path = os.path.join(save_dir, f"{dataset_name}_aliases.txt")
    with open(file_path, "w", encoding="utf-8") as file:
        for cui, entity in ontology.entities.items():
            file.write(f"{cui}||{entity.name}\n")
            if len(entity.aliases) != 0:
                if isinstance(entity.aliases, list):
                    for word in entity.aliases:
                        if word is not None:
                            # print(word)
                            word = word.strip()
                            file.write(f"{cui}||{word}\n")
                else:
                    # words = entity.aliases.split('; , ')
                    words = re.split("[;|]", entity.aliases)
                    for word in words:
                        word = word.strip()
                        file.write(f"{cui}||{word}\n")
            if entity.equivalant_cuis:
                for eqcui in entity.equivalant_cuis:
                    if eqcui != entity.cui:
                        file.write(f"{eqcui}||{entity.name}\n")
                        if entity.aliases:
                            words = re.split("[;|]", entity.aliases)
                            for word in words:
                                word = word.strip()
                                file.write(f"{eqcui}||{word}\n")

    return file_path


def create_tfidf_ann_index(out_path: str, input_path_kb: str):
    """
    Build tfidf vectorizer and ann index.

    Parameters
    ----------
    out_path: str, required.
        The path where the various model pieces will be saved.
    input_path_kb : KnowledgeBase, required
        The kb items to generate the index and vectors for.

    """
    tfidf_vectorizer_path = f"{out_path}/tfidf_vectorizer.joblib"
    concept_aliases = []
    os.makedirs(out_path, exist_ok=True)

    # Open the text file for reading
    with open(input_path_kb, "r") as file:
        # Read each line from the file
        for line in file:
            # Split the line at "||" and extract the word after the separator
            parts = line.strip().split("||")
            if len(parts) > 1:
                alias = parts[1]
                # Append the alias to the list
                concept_aliases.append(alias)

    # concept_aliases = list(kb.alias_to_cuis.keys())

    # NOTE: here we are creating the tf-idf vectorizer with float32 type, but we can serialize the
    # resulting vectors using float16, meaning they take up half the memory on disk. Unfortunately
    # we can't use the float16 format to actually run the vectorizer, because of this bug in sparse
    # matrix representations in scipy: https://github.com/scipy/scipy/issues/7408
    print(f"Fitting tfidf vectorizer on {len(concept_aliases)} aliases")
    tfidf_vectorizer = TfidfVectorizer(
        analyzer="char_wb", ngram_range=(3, 3), min_df=10, dtype=np.float32
    )
    start_time = datetime.datetime.now()
    concept_alias_tfidfs = tfidf_vectorizer.fit_transform(concept_aliases)
    print(f"Saving tfidf vectorizer to {tfidf_vectorizer_path}")
    joblib.dump(tfidf_vectorizer, tfidf_vectorizer_path)
    end_time = datetime.datetime.now()
    total_time = end_time - start_time
    print(f"Fitting and saving vectorizer took {total_time.total_seconds()} seconds")


def put_mention_in_context(
    doc_id,
    doc,
    mention,
    offsets,
    max_len,
    start_delimiter="START ",
    end_delimiter=" END",
    resolve_abbrevs=False,
    abbrev_dict=None,
):
    """
    Put a mention in context of surrouding text
    """
    # Resolve abbreviaions if desired
    if resolve_abbrevs and abbrev_dict is not None:
        mention = resolve_abbreviation(doc_id, mention, abbrev_dict)

    tagged_mention = start_delimiter + mention + end_delimiter
    start = offsets[0][0]
    end = offsets[-1][-1]
    before_context = doc[:start]
    after_context = doc[end:]
    before_split_context = before_context.split(" ")
    after_split_context = after_context.split(" ")
    len_before = len(before_split_context)
    len_after = len(after_split_context)

    max_ctx_l_len = max_len // 2
    max_ctx_r_len = max_len - max_ctx_l_len

    if len_before < max_ctx_l_len and len_after < max_ctx_r_len:
        mention_in_context = before_context + tagged_mention + after_context

    elif len_before >= max_ctx_l_len and len_after >= max_ctx_r_len:
        mention_in_context = (
            " ".join(before_split_context[-max_ctx_l_len:])
            + " "
            + tagged_mention
            + " "
            + " ".join(after_split_context[:max_ctx_r_len])
        )

    elif len_before >= max_ctx_l_len:
        ctx_l_len = max_len - len_after
        mention_in_context = (
            " ".join(before_split_context[-ctx_l_len:])
            + " "
            + tagged_mention
            + after_context
        )
    else:
        ctx_r_len = max_len - len_before
        mention_in_context = (
            before_context
            + tagged_mention
            + " "
            + " ".join(after_split_context[:ctx_r_len])
        )

    return mention_in_context


def contextualize_mentions(
    doc_dict,
    deduplicated,
    max_len=128,
    resolve_abbrevs=False,
    abbrev_dict=None,
    start_delimiter="START ",
    end_delimiter=" END",
):
    deduplicated["contextualized_mention"] = deduplicated[
        ["document_id", "mention", "offsets"]
    ].progress_apply(
        lambda x: put_mention_in_context(
            x[0],
            doc_dict[x[0]],
            x[1],
            x[2],
            max_len=max_len,
            resolve_abbrevs=resolve_abbrevs,
            abbrev_dict=abbrev_dict,
        ),
        axis=1,
    )

    return deduplicated


def create_training_files(
    save_dir: str,
    dataset_name: str,
    document_dict: dict,
    deduplicated,
    abbreviations_dict: dict,
    cui2alias: dict,
    resolve_abbrevs: bool = True,
):
    # Get contextualized mentions
    # print("deduplicated here", deduplicated)
    test_docs = deduplicated[deduplicated["split"] == "validation"]
    print(test_docs)

    df = contextualize_mentions(
        document_dict,
        deduplicated,
        resolve_abbrevs=resolve_abbrevs,
        abbrev_dict=abbreviations_dict,
    )

    # df["entity_aliases"] = df["db_ids"].map(
    #     lambda x: list(set([z for y in x for z in cui2alias[y]]))
    # )
    # df["most_similar_alias"] = df[["mention", "entity_aliases"]].apply(
    #     lambda x: get_most_similar_alias(x[0], x[1], vectorizer)
    # )

    # Get closest synonym for each mention
    tfidf_vectorizer = f"file_dumps/{dataset_name}/tfidf_vectorizer.joblib"
    vectorizer = joblib.load(tfidf_vectorizer)

    # df["entity_aliases"] = df["db_ids"].map(
    #     lambda x: list(set([z for y in x for z in cui2alias[y]]))
    # )
    def get_aliases(db_ids):
        aliases = []
        for db_id in db_ids:
            if db_id in cui2alias:
                aliases.extend(cui2alias[db_id])
        return list(set(aliases))

    df["entity_aliases"] = df["db_ids"].map(get_aliases)

    # print(df[df.entity_aliases.map(lambda x: len(x)==0)])
    print("Getting most similar alias")
    df["most_similar_alias"] = df[["mention", "entity_aliases"]].progress_apply(
        lambda x: get_most_similar_alias(
            mention=x[0], cui_alias_list=x[1], vectorizer=vectorizer
        ),
        axis=1,
    )

    # Make json string in correct format for decoder
    df["source_json"] = df["contextualized_mention"].map(lambda x: ujson.dumps([x]))

    df["target_json"] = df[["mention", "most_similar_alias"]].progress_apply(
        lambda x: ujson.dumps([f"{x[0]} is", x[1] if x[1] is not None else "unknown"]),
        axis=1,
    )

    # Store/check results
    # print(df.head(5))
    # print("Saving Pickle)")
    # df.to_pickle(os.path.join(save_dir, "processed_mentions.pickle"))

    # Write data files to pickle
    for split in df.split.unique():
        subset = df.query("split == @split")
        split_name = split
        if split in ["valid", "validation"]:
            split_name = "dev"

        # Write files
        with open(os.path.join(save_dir, f"{split_name}label.txt"), "w") as f:
            entity_link_list = ["|".join(x) for x in subset.db_ids.tolist()]
            f.write("\n".join(entity_link_list))

        with open(os.path.join(save_dir, f"{split_name}.source"), "w") as f:
            f.write("\n".join(subset.source_json.tolist()))

        with open(os.path.join(save_dir, f"{split_name}.target"), "w") as f:
            f.write("\n".join(subset.target_json.tolist()))


def get_most_similar_alias(mention, cui_alias_list, vectorizer):
    """
    Get most similar CUI alias to current mention using TF-IDF.
    Returns None if cui_alias_list is empty.
    """
    if not cui_alias_list:
        return None
    most_similar_idx = cal_similarity_tfidf(cui_alias_list, mention, vectorizer)[0]
    return cui_alias_list[most_similar_idx]


def cal_similarity_tfidf(a: list, b: str, vectorizer):
    features_a = vectorizer.transform(a)
    features_b = vectorizer.transform([b])
    features_T = features_a.T
    sim = features_b.dot(features_T).todense()
    return sim[0].argmax(), np.max(np.array(sim)[0])


def create_target_kb_dict(file_path, dataset_name):
    with open(file_path + f"{dataset_name}_aliases.txt", "r") as f:
        lines = f.read().split("\n")

    # Construct dict mapping each CURIE to a list of aliases
    umls_dict = {}
    for line in tqdm(lines):
        if len(line.split("||")) != 2:
            print(line)
            continue
        cui, name = line.split("||")
        if cui in umls_dict:
            umls_dict[cui].add(name.lower())
        else:
            umls_dict[cui] = set([name.lower()])

    processed_dict = {key: list(val) for key, val in umls_dict.items()}
    return processed_dict


def data_preprocess(
    dataset_name: str,
    save_dir: str,
    ontology: BiomedicalOntology,
    path_to_abbrev: str,
    resolve_abbrevs=False,
):

    if not resolve_abbrevs:
        save_dir = os.path.join(save_dir, "no_abbr_res/")
    else:
        save_dir = os.path.join(save_dir, "abbr_res/")

    save_dir = os.path.join(save_dir, f"{dataset_name}/")

    mappings_dir = cuis_to_aliases(ontology, save_dir, dataset_name)
    create_tfidf_ann_index(f"file_dumps/{dataset_name}", mappings_dir)

    # read data
    # dataset = load_dataset(f"bigbio/{dataset_name}", name=f"{dataset_name}_bigbio_kb")
    dataset = load_bigbio_dataset(f"{dataset_name}")

    target_kb_dict = create_target_kb_dict(save_dir, dataset_name)
    # create target_kb.json
    with open(os.path.join(save_dir, "target_kb.json"), "w") as target_kb:
        str_target = ujson.dumps(target_kb_dict, indent=2)
        target_kb.write(str_target)

    if resolve_abbrevs:
        with open(path_to_abbrev) as json_file:
            abbreviations_dict = ujson.load(json_file)
    else:
        abbreviations_dict = None

    entity_remapping_dict = CUIS_TO_REMAP[dataset_name]
    entities_to_exclude = CUIS_TO_EXCLUDE[dataset_name]

    if dataset_name in VALIDATION_DOCUMENT_IDS:
        validation_pmids = VALIDATION_DOCUMENT_IDS[dataset_name]
    else:
        validation_pmids = None

    deduplicated = dataset_to_df(
        dataset,
        entity_remapping_dict=entity_remapping_dict,
        cuis_to_exclude=entities_to_exclude,
        val_split_ids=validation_pmids,
    )
    deduplicated["mention"] = deduplicated["text"]

    for split in deduplicated.split.unique():
        print(split)
    split_docs = dataset_to_documents(dataset)

    # create source file
    create_training_files(
        save_dir,
        dataset_name=dataset_name,
        document_dict=split_docs,
        deduplicated=deduplicated,
        abbreviations_dict=abbreviations_dict,
        cui2alias=target_kb_dict,
        resolve_abbrevs=resolve_abbrevs,
    )

    return save_dir
